#!/usr/bin/perl

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use warnings;

use Config4GNU::DNS::ZoneFileParser;
use Config4GNU::ISC;

=head1 NAME

Config4GNU::DNS::Bind - a Perl module for accessing Bind DNS data

=head1 DESCRIPTION

=cut

package Config4GNU::DNS::Bind;

=head2 Config4GNU::DNS::Bind

=head3 Methods

=over

=item my $bind = Config4GNU::DNS::Bind->new( 'NAMED_CONF_FILE' => $filename );

Create a new object used for managing Bind configuration and zone data.
You must specify the location of the named.conf file.

=cut

sub new
{
	my $class = shift;
	my %args = @_;

	my $self = {
		'NAMED_CONF_FILE' => $args{'NAMED_CONF_FILE'}
	};
	bless $self, $class;
	
	$self->init;
	return $self;
}

# internal method
sub init
{
	my $self = shift;
	
	$self->{'NAMED_CONF'} = Config4GNU::ISC->new(
			$self->{'NAMED_CONF_FILE'});
}

=item my @zones = $bind->zones;

Returns an array of BindZone objects, one for each zone defined in the
named.conf file.

=cut

sub zones
{
	my $self = shift;
	my @result = ();
	
	foreach my $section (@{$self->{'NAMED_CONF'}->children})
	{
		if ($section->word(0) eq 'zone')
		{
			my $origin = $section->word(1);
			my $type = $section->getPropertyValue('type');
			my $file = $section->getPropertyValue('file');
			if ($type eq 'master')
			{
				my $zone = Config4GNU::DNS::BindZone->new(
					'FILENAME' => $file,
					'ORIGIN' => $origin);
				push @result, $zone;
			}
		}
	}
	return @result;
}

=item my $zone = $bind->getZone('example.org');

Returns the specified zone, or C<undef> if the zone is not found.

=cut

sub getZone
{
	my $self = shift;
	my ($zonename) = @_;

	foreach my $zone ($self->zones)
	{
		return $zone if ($zone->name eq $zonename);
	}
	return undef;
}

=back

=cut

package Config4GNU::DNS::BindZone;
use base 'Config4GNU::DNS::ZoneFileParser';

=head2 Config4GNU::DNS::BindZone

=head3 Methods

=over

=item my $name = $zone->name;

Returns the zone's name (or origin) as a string.

=cut

sub name
{
	my $self = shift;
	return $self->{'ORIGIN'};
}

=item my $filename = $zone->file;

Returns the name of the zone's data file.

=cut

sub file
{
	my $self = shift;
	return $self->{'FILENAME'};
}

=item my $type = $zone->type;

For now, always returns 'master'. Future implementations may also return
'slave', 'hints', or 'cache'. Future implementations may also replace the
'master' and 'slave' terms with 'primary' and 'secondary'.

=cut

sub type
{
	return 'master';
}

=item my $class = $zone->class;

For now, always returns 'IN', since this is the only supported domain
class.

=cut

sub class
{
	return 'IN';
}

=item my @records = $zone->records;

Returns an array of BindResourceRecord objects, one for each resource
record found in the zone file.

=cut

sub records
{
	my $self = shift;

	my %unique_counts = ();
	my @result = ();
	foreach my $entry (@{$self->{'CHILDREN'}})
	{
		if ($entry->isa('Config4GNU::DNS::ZoneFileParser::ResourceRecord'))
		{
			my $record = Config4GNU::DNS::BindResourceRecord::create($entry);
			my $hash = $record->name . " " . $record->class
				. " " . $record->type;
			$record->{'ID'} = ++$unique_counts{$hash};
			push @result, $record;
		}
	}
	return @result;
}

sub findRecord
{
	my $self = shift;
	my %keys = @_;

	foreach my $record ($self->records)
	{
		next if (defined($keys{'OWNER'})
			&& $keys{'OWNER'} ne $record->name);
		next if (defined($keys{'CLASS'})
			&& $keys{'CLASS'} ne $record->class);
		next if (defined($keys{'TYPE'})
			&& $keys{'TYPE'} ne $record->type);
		next if (defined($keys{'ID'})
			&& $keys{'ID'} ne $record->{'ID'});

		return $record;
	}
	return undef;
}

=back

=cut

package Config4GNU::DNS::BindResourceRecord;

=head2 Config4GNU::DNS::BindResourceRecord

=head3 Methods

=over

=item my $record = Config4GNU::DNS::BindResourceRecord::create($entry);

$entry is of type Config4GNU::DNS::ZoneFileParser::ResourceRecord.
Returns a wrapper object of a class dependent on what type of record
$entry contains.

=cut

sub create
{
	my ($entry) = @_;

	my $self = { 'ENTRY' => $entry };
	if ($entry->type eq 'A')
	{
		bless $self, 'Config4GNU::DNS::BindARecord';
	}
	elsif ($entry->type eq 'SOA')
	{
		bless $self, 'Config4GNU::DNS::BindSOARecord';
	}
	elsif ($entry->type eq 'MX')
	{
		bless $self, 'Config4GNU::DNS::BindMXRecord';
	}
	else
	{
		bless $self, 'Config4GNU::DNS::BindResourceRecord';
	}
	return $self;
}

=item my $name = $record->name;

Returns the domain name that owns the given record $record.

=cut

sub name
{
	my $self = shift;
	return $self->{'ENTRY'}->domainName;
}

=item my $ttl = $record->ttl;

Returns the time-to-live (TTL) of the resource record, specified in
seconds.

=cut

sub ttl
{
	my $self = shift;
	return $self->{'ENTRY'}->ttl;
}

=item my $class = $record->class;

Returns the resource record's class (e.g. 'IN').

=cut

sub class
{
	my $self = shift;
	return $self->{'ENTRY'}->class;
}

=item my $type = $record->type;

Returns the resource record's type as a string (e.g. 'A', 'MX', 'NS').

=cut

sub type
{
	my $self = shift;
	return $self->{'ENTRY'}->type;
}

=item my $rdata = $record->rdata;

Returns the resource record's record data as a string. The record data
varies depending on the record's type.

=cut

sub rdata
{
	my $self = shift;
	return $self->{'ENTRY'}->rdataAsString;
}

# internal function
# takes a domain name and fully-qualifies it, removing the trailing
# dot when finished
sub fqdn
{
	my $self = shift;
	my ($domainName) = @_;

	unless ($domainName =~ /\.$/)
	{
		# add origin
		$domainName .= '.' . $self->{'ENTRY'}->{'ORIGIN'};
	}
	$domainName =~ s/\.$//;
	return $domainName;
}

# internal function
# takes an interval expression as found in zone files and converts
# it to a number expressed in seconds
sub parseduration
{
	my $self = shift;
	my $duration = shift;

	my $SECONDS_PER_MINUTE = 60;
	my $SECONDS_PER_HOUR = 60 * $SECONDS_PER_MINUTE;
	my $SECONDS_PER_DAY = 24 * $SECONDS_PER_HOUR;
	my $SECONDS_PER_WEEK = 7 * $SECONDS_PER_DAY;

	my $seconds = 0;
	if ($duration =~ /^(\d+)w/)
	{
		$seconds += ($1 * $SECONDS_PER_WEEK);
		$duration = $';
	}
	if ($duration =~ /^(\d+)d/)
	{
		$seconds += ($1 * $SECONDS_PER_DAY);
		$duration = $';
	}
	if ($duration =~ /^(\d+)h/)
	{
		$seconds += ($1 * $SECONDS_PER_HOUR);
		$duration = $';
	}
	if ($duration =~ /^(\d+)m/)
	{
		$seconds += ($1 * $SECONDS_PER_MINUTE);
		$duration = $';
	}
	if ($duration =~ /^(\d+)s?$/)
	{
		$seconds += $1;
	}
	elsif ($duration ne '')
	{
		warn "unrecognized TTL $duration\n";
	}

	return $seconds;
}

=back

=cut

package Config4GNU::DNS::BindARecord;
use base 'Config4GNU::DNS::BindResourceRecord';

=head2 Config4GNU::DNS::BindARecord

This is a subclass of Config4GNU::DNS::BindResourceRecord.
It implements functionality unique to IPv4 Address (A) records.

=head3 Methods

=over

=item my $ip = $record->address;

Returns the address specified by the record.

=cut

sub address
{
	my $self = shift;
	return $self->rdata;
}

=back

=cut

package Config4GNU::DNS::BindMXRecord;
use base 'Config4GNU::DNS::BindResourceRecord';

=head2 Config4GNU::DNS::BindMXRecord

This is a subclass of Config4GNU::DNS::BindResourceRecord.
It implements functionality unique to mail exchanger (MX) records.

=head3 Methods

=over

=item my $mxhost = $record->mxhost;

Returns the fully-qualified domain name specified by the MX record
as the mail exchanger for the owner domain.

=cut

sub mxhost
{
	my $self = shift;
	my ($pref, $mxhost) = $self->{'ENTRY'}->rdataAsWords;
	return $self->fqdn($mxhost);
}

=item my $pref = $record->preference

Returns the preference (an integer) specified by the resource record.

=cut

sub preference
{
	my $self = shift;
	my ($pref, $mxhost) = $self->{'ENTRY'}->rdataAsWords;
	return $pref;
}

=item my $rdata = $record->rdata;

Overrides method of base class Config4GNU::DNS::BindResourceRecord.
Returns record data in the format "[$pref] $mxhost".

=cut

sub rdata
{
	my $self = shift;
	my $pref = $self->preference;
	my $mxhost = $self->mxhost;
	return "[$pref] $mxhost";
}

=back

=cut

package Config4GNU::DNS::BindSOARecord;
use base 'Config4GNU::DNS::BindResourceRecord';

=head2 Config4GNU::DNS::BindSOARecord

This is a subclass of Config4GNU::DNS::BindResourceRecord.
It implements functionality unique to start-of-authority (SOA) records.

=head3 Methods

=over

=item my ($mname, $rname, $serial, $refresh, $retry, $expire, $negcache) = $record->getParts

Returns the components of the SOA record as a list.

=cut

sub getParts
{
	my $self = shift;
	my ($mname, $rname, $serial, $refresh, $retry,
		$expire, $negcache) = $self->{'ENTRY'}->rdataAsWords;
	$mname = $self->fqdn($mname);
	$rname = $self->fqdn($rname);
	$refresh = $self->parseduration($refresh);
	$retry = $self->parseduration($retry);
	$expire = $self->parseduration($expire);
	$negcache = $self->parseduration($negcache);

	return ($mname, $rname, $serial, $refresh, $retry,
		$expire, $negcache);
}

=item my $rdata = $record->rdata;

Overrides method of base class Config4GNU::DNS::BindResourceRecord.
Returns record data in the format "[$serial] $mname, $rname".

=cut

sub rdata
{
	my $self = shift;
	my @words = $self->{'ENTRY'}->rdataAsWords;
	my ($mname, $rname, $serial, $refresh, $retry,
		$expire, $negcache) = @words;
	
	return "[$serial] " . $self->fqdn($mname) . ", " . $self->fqdn($rname);
}

=back

=cut

1;

=head1 AUTHOR

Jason Long <jason@long.name>

=cut
