#!/usr/bin/perl

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# USAGE:
#   Import the associated schema file (bind.mof) into the OpenWBEM
#   CIMOM using owmofc.
#   Install this file to $(OPENWBEM_PREFIX)/lib/openwbem/perlproviders/

use strict;
use warnings;
use Config4GNU::WBEM;
use Config4GNU::DNS::Bind;

use constant CLASSNAME_DNSSERVER => 'Config4GNU_Bind_DNSServer';
use constant CLASSNAME_DNSZONE => 'Config4GNU_Bind_DNSZone';
use constant CLASSNAME_ATYPE => 'Config4GNU_Bind_AType';
use constant CLASSNAME_CNAMETYPE => 'Config4GNU_Bind_CNAMEType';
use constant CLASSNAME_MXTYPE => 'Config4GNU_Bind_MXType';
use constant CLASSNAME_NSTYPE => 'Config4GNU_Bind_NSType';
use constant CLASSNAME_PTRTYPE => 'Config4GNU_Bind_PTRType';
use constant CLASSNAME_SOATYPE => 'Config4GNU_Bind_SOAType';
use constant CLASSNAME_SRVTYPE => 'Config4GNU_Bind_SRVType';
use constant CLASSNAME_TXTTYPE => 'Config4GNU_Bind_TXTType';
use constant CLASSNAME_OTHERTYPE => 'Config4GNU_Bind_OtherType';
use constant CLASSNAME_HOSTEDDNSSERVICE
	=> 'Config4GNU_Bind_HostedDNSService';
use constant CLASSNAME_SERVERZONECONTAINMENT
	=> 'Config4GNU_Bind_ServerZoneContainment';
use constant CLASSNAME_ZONERESOURCERECORDCONTAINMENT
	=> 'Config4GNU_Bind_ZoneResourceRecordContainment';

use constant DEFAULT_CONFIGFILE => '/etc/named.conf';

# initialize standard providers
Server->register(CLASSNAME_DNSSERVER);
Zone->register(CLASSNAME_DNSZONE);
ResourceRecord->register(CLASSNAME_ATYPE);
ResourceRecord->register(CLASSNAME_CNAMETYPE);
ResourceRecord->register(CLASSNAME_MXTYPE);
ResourceRecord->register(CLASSNAME_NSTYPE);
ResourceRecord->register(CLASSNAME_PTRTYPE);
ResourceRecord->register(CLASSNAME_SOATYPE);
ResourceRecord->register(CLASSNAME_SRVTYPE);
ResourceRecord->register(CLASSNAME_TXTTYPE);
ResourceRecord->register(CLASSNAME_OTHERTYPE);

# initialize association class providers
HostedService->register(CLASSNAME_HOSTEDDNSSERVICE);
ServerZoneContainment->register(CLASSNAME_SERVERZONECONTAINMENT);
ZoneResourceRecordContainment->register(CLASSNAME_ZONERESOURCERECORDCONTAINMENT);

sub getBind
{
	my $configfile = shift;
	my $bind = Config4GNU::DNS::Bind->new('NAMED_CONF_FILE' => $configfile);
	return $bind;
}

#######################################################################
#######################################################################

package Server;
use base 'Config4GNU::WBEM::SimpleProvider';
use constant CLASSNAME_DNSSERVER => main::CLASSNAME_DNSSERVER;

sub enumInstanceNames
{
	my $self = shift;
	my ($env, $ns, $className, $result, $cimClass) = @_;

	my $configfile = main::DEFAULT_CONFIGFILE;
	
	my $mcop = owperl::CIMObjectPath->new(CLASSNAME_DNSSERVER, $ns);
	chomp (my $hostname = `hostname`);
	$mcop->addStringKeyValue('SystemName', $hostname);
	$mcop->addStringKeyValue('SystemCreationClassName', 'CIM_System');
	$mcop->addStringKeyValue('Name', $configfile);
	$mcop->addStringKeyValue('CreationClassName', CLASSNAME_DNSSERVER);

	$result->handle($mcop);
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $configfile = $instanceName->getStringKeyValue('Name');

	my $ci = $cimClass->newInstance;

	# fill in the key properties
	$ci->setStringProperty('SystemName',
		$instanceName->getStringKeyValue('SystemName'));
	$ci->setStringProperty('SystemCreationClassName',
		$instanceName->getStringKeyValue('SystemCreationClassName'));
	$ci->setStringProperty('Name', $configfile);
	$ci->setStringProperty('CreationClassName', CLASSNAME_DNSSERVER);

	# other properties...
	$ci->setStringProperty('ElementName', 'BIND DNS Server');

	return $ci;
}

#######################################################################
#######################################################################

package HostedService;
use base 'Config4GNU::WBEM::SimpleProvider';

sub associatorNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole) = @_;
	
	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_DNSSERVER))
	{
		my $mcop = owperl::CIMObjectPath->new(
			$objectName->getStringKeyValue('SystemCreationClassName'), $ns);
		$mcop->addStringKeyValue('Name',
			$objectName->getStringKeyValue('SystemName'));
		$mcop->addStringKeyValue('CreationClassName',
			$objectName->getStringKeyValue('SystemCreationClassName'));
		$result->handle($mcop);	
	}
	elsif ($provider->is_subclass($env, $ns, $baseClass, 'CIM_System'))
	{
		my $configfile = main::DEFAULT_CONFIGFILE;
		
		my $mcop = owperl::CIMObjectPath->new(main::CLASSNAME_DNSSERVER, $ns);
		$mcop->addStringKeyValue('Name', $configfile);
		$mcop->addStringKeyValue('CreationClassName', main::CLASSNAME_DNSSERVER);
		$mcop->addStringKeyValue('SystemName',
			$objectName->getStringKeyValue('Name'));
		$mcop->addStringKeyValue('SystemCreationClassName',
			$objectName->getStringKeyValue('CreationClassName'));
		$result->handle($mcop);
	}
}

sub referenceNames
{
}

#######################################################################
#######################################################################

package Zone;
use base 'Config4GNU::WBEM::SimpleProvider';
use constant CLASSNAME_DNSZONE => main::CLASSNAME_DNSZONE;

sub enumInstanceNames
{
	my $self = shift;
	my ($env, $ns, $className, $result, $cimClass) = @_;

	my $bind = main::getBind(main::DEFAULT_CONFIGFILE);

	foreach my $zone ($bind->zones)
	{
		my $mcop = owperl::CIMObjectPath->new(CLASSNAME_DNSZONE, $ns);
		chomp (my $hostname = `hostname`);
		$mcop->addStringKeyValue('DnsServerName', $hostname);
		$mcop->addStringKeyValue('Name', $zone->name);
		$result->handle($mcop);
	}
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $bind = main::getBind(main::DEFAULT_CONFIGFILE);
	my $zone = $bind->getZone($instanceName->getStringKeyValue('Name'));

	my $ci = $cimClass->newInstance;

	# fill in the key properties
	$ci->setStringProperty('DnsServerName',
		$instanceName->getStringKeyValue('DnsServerName'));
	$ci->setStringProperty('Name', $zone->name);

	# other properties...
	$ci->setStringProperty('File', $zone->file);
	$ci->setStringProperty('Type', $zone->type);
	$ci->setStringProperty('Class', $zone->class);
	$ci->setStringProperty('ElementName', $zone->name);

	return $ci;
}

#######################################################################
#######################################################################

package ResourceRecord;
use base 'Config4GNU::WBEM::SimpleProvider';

my %types = (
	'A' => main::CLASSNAME_ATYPE,
	'CNAME' => main::CLASSNAME_CNAMETYPE,
	'MX' => main::CLASSNAME_MXTYPE,
	'NS' => main::CLASSNAME_NSTYPE,
	'PTR' => main::CLASSNAME_PTRTYPE,
	'SOA' => main::CLASSNAME_SOATYPE,
	'SRV' => main::CLASSNAME_SRVTYPE,
	'TXT' => main::CLASSNAME_TXTTYPE,
	);

sub enumInstanceNames
{
	my $self = shift;
	my ($env, $ns, $className, $result, $cimClass) = @_;

	my $bind = main::getBind(main::DEFAULT_CONFIGFILE);
	chomp (my $hostname = `hostname`);

	foreach my $zone ($bind->zones)
	{
		foreach my $record ($zone->records)
		{
			if (defined $types{$record->type})
			{
				next unless (lc($className) eq lc($types{$record->type}));
			}
			else
			{
				next unless (lc($className) eq lc(main::CLASSNAME_OTHERTYPE));
			}
			my $mcop = owperl::CIMObjectPath->new($className, $ns);
			$mcop->addStringKeyValue('DnsServerName', $hostname);
			$mcop->addStringKeyValue('ZoneName', $zone->name);
			record2cop($mcop, $record);
			$result->handle($mcop);
		}
	}
}

sub record2cop
{
	my ($mcop, $record) = @_;
	$mcop->addStringKeyValue('OwnerName', $record->name);
	$mcop->addStringKeyValue('Class', $record->class);
	my $type = $record->type;
	$mcop->addStringKeyValue('Type', $type);
	$mcop->addIntegerKeyValue('idNumber', $record->{'ID'});
	return $mcop;
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $bind = main::getBind(main::DEFAULT_CONFIGFILE);
	my $zone = $bind->getZone($instanceName->getStringKeyValue('ZoneName'));
	my $record = $zone->findRecord(
			OWNER => $instanceName->getStringKeyValue('OwnerName'),
			CLASS => $instanceName->getStringKeyValue('Class'),
			TYPE => $instanceName->getStringKeyValue('Type'),
			ID => $instanceName->getIntegerKeyValue('idNumber'));
	if (not defined $record)
	{
		die "Record not found.\n";
	}

	my $ci = $cimClass->newInstance;

	# fill in the key properties
	$ci->setStringProperty('DnsServerName',
		$instanceName->getStringKeyValue('DnsServerName'));
	$ci->setStringProperty('ZoneName', $zone->name);
	$ci->setStringProperty('OwnerName', $record->name);
	$ci->setStringProperty('Class', $record->class);
	my $type = $record->type;
	$ci->setStringProperty('Type', $type);
	$ci->setIntegerProperty('idNumber', $record->{'ID'});

	# other properties...
	$ci->setIntegerProperty('TTL', $record->ttl);
	$ci->setStringProperty('RecordData', $record->rdata);
	$ci->setStringProperty('ElementName', $record->name);

	if ($type eq 'A' && lc($types{$type}) eq lc($cimClass->getName)
		&& $record->isa('Config4GNU::DNS::BindARecord'))
	{
		# set A-type additional properties
		$ci->setStringProperty('IPAddress', $record->address);
	}
	elsif ($type eq 'MX' && lc($types{$type}) eq lc($cimClass->getName)
		&& $record->isa('Config4GNU::DNS::BindMXRecord'))
	{
		# set MX-type additional properties
		$ci->setStringProperty('MailExchange', $record->mxhost);
		$ci->setIntegerProperty('Preference', $record->preference);
	}
	elsif ($type eq 'SOA' && lc($types{$type}) eq lc($cimClass->getName)
		&& $record->isa('Config4GNU::DNS::BindSOARecord'))
	{
		# set SOA-type additional properties
		my ($mname, $rname, $serial, $refresh, $retry, $expire, $negcache)
			= $record->getParts;
		$ci->setStringProperty('PrimaryServer', $mname);
		$ci->setStringProperty('ResponsibleParty', $rname);
		$ci->setIntegerProperty('SerialNumber', $serial);
		$ci->setIntegerProperty('RefreshInterval', $refresh);
		$ci->setIntegerProperty('RetryDelay', $retry);
		$ci->setIntegerProperty('ExpireLimit', $expire);
		$ci->setIntegerProperty('NegCache', $negcache);
	}

	return $ci;
}

#######################################################################
#######################################################################

package ServerZoneContainment;
use base 'Config4GNU::WBEM::SimpleProvider';

sub associatorNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole) = @_;

	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_DNSSERVER))
	{
		my $bind = main::getBind(main::DEFAULT_CONFIGFILE);
		foreach my $zone ($bind->zones)
		{
			my $mcop = owperl::CIMObjectPath->new(main::CLASSNAME_DNSZONE, $ns);
			chomp (my $hostname = `hostname`);
			$mcop->addStringKeyValue('DnsServerName', $hostname);
			$mcop->addStringKeyValue('Name', $zone->name);
			$result->handle($mcop);
		}
	}
}

#######################################################################
#######################################################################

package ZoneResourceRecordContainment;
use base 'Config4GNU::WBEM::SimpleProvider';

sub associatorNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole) = @_;

	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_DNSZONE))
	{
		my $bind = main::getBind(main::DEFAULT_CONFIGFILE);
		my $zone = $bind->getZone($objectName->getStringKeyValue('Name'));

		foreach my $record ($zone->records)
		{
			my $className = $types{$record->type};
			unless (defined $className)
			{
				$className = main::CLASSNAME_OTHERTYPE;
			}

			my $mcop = owperl::CIMObjectPath->new($className, $ns);
			$mcop->addStringKeyValue('DnsServerName',
				$objectName->getStringKeyValue('DnsServerName'));
			$mcop->addStringKeyValue('ZoneName', $zone->name);
			ResourceRecord::record2cop($mcop, $record);
			$result->handle($mcop);
		}
	}
}
