#!/usr/bin/perl

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use warnings;

use Config4GNU::Tokenizer;

=head1 NAME

Config4GNU::DNS::ZoneFileParser - read/write BIND master zone files

=head1 DESCRIPTION

=cut

package Config4GNU::DNS::ZoneFileParser;

=head2 Config4GNU::DNS::ZoneFileParser

=head3 Methods

=over

=item my $parser = Config4GNU::DNS::ZoneFileParser->new(
	'FILENAME' => $filename, 'ORIGIN' => $origin );

Create a new parser, interpret the specified file as a zone file,
and return an object that can be used to access/modify the contents
of the zone file.

=cut

sub new
{
	my $class = shift;
	my %args = @_;

	my $self = {
		'FILENAME' => $args{'FILENAME'},
		'ORIGIN' => $args{'ORIGIN'}
	};
	bless $self, $class;

	$self->init;
	return $self;
}

=item $parser->save;

Writes out the configuration file, complete with any changes made.

=cut

sub save
{
	my $self = shift;
	
	my $fh;
	open $fh, '>', $self->filename
		or die "Cannot save: $!\n";

	foreach my $entry (@{$self->{'CHILDREN'}})
	{
		print $fh $entry->toString;
	}
	close $fh;
}

# internal method
sub init
{
	my $self = shift;

	# create a tokenizer
	my $reader = Config4GNU::Tokenizer->openFile($self->filename);

	$self->{'CHILDREN'} = [];

	my $entry;
	my $comment = '';
	my $currentDomainName;
	my $currentTtl = '86400';
	my $currentClass = 'IN';
	while (defined($entry = nextEntry($reader)))
	{
		if ($entry->isComment)
		{
			$comment .= $entry->toString;
		}
		elsif ($entry->token(0) =~ /^\$/)
		{
			# a command entry
			my $command = Config4GNU::DNS::ZoneFileParser::CommandEntry::create(
					$entry);
			$command->setComment($comment);
			$comment = '';

			# commands change the meaning of the following entries
			if ($command->isa('Config4GNU::DNS::ZoneFileParser::TtlCommandEntry'))
			{
				$currentTtl = $command->getTtl;
			}

			$self->add($command);
		}
		else
		{
			# a resource record
			my $record = Config4GNU::DNS::ZoneFileParser::ResourceRecord->new(
					$entry);
			$record->setOrigin($self->{'ORIGIN'});
			$record->setComment($comment);
			$record->setCurrentDomainName($currentDomainName);
			$record->setCurrentTtl($currentTtl);
			$record->setCurrentClass($currentClass);
			$comment = '';
			$currentDomainName = $record->domainName;

			$self->add($record);
		}
	}
}

sub debug
{
	my $self = shift;

	foreach my $entry (@{$self->{'CHILDREN'}})
	{
		if ($entry->isa('Config4GNU::DNS::ZoneFileParser::CommandEntry'))
		{
			print "Command entry:\n";
			print $entry->data;
		}
		elsif ($entry->isa('Config4GNU::DNS::ZoneFileParser::ResourceRecord'))
		{
			print "Resource record for ", $entry->domainName, ":\n";
			print "  class: ", $entry->class, "\n";
			print "  TTL:   ", $entry->ttl, "\n";
			print "  type:  ", $entry->type, "\n";
			print "  rdata: ", $entry->rdataAsString, "\n";
		}
	}
}

# internal method
sub add
{
	my $self = shift;
	my $entry = shift;

	if ($entry->isa('Config4GNU::DNS::ZoneFileParser::EntryBase'))
	{
		push @{$self->{'CHILDREN'}}, $entry;
	}
	else
	{
		warn "Cannot add type " . ref($entry);
	}
}

# internal method
sub nextEntry
{
	my $tokenizer = shift;

	# get the whole entry
	my @entry_tokens = ();
	my $paren = 0;
	my $token;
	while (defined($token = $tokenizer->nextToken))
	{
		push @entry_tokens, $token;
		if ($token eq '(')
		{
			$paren++;
		}
		elsif ($token eq ')')
		{
			$paren--;
		}
		elsif ($token =~ /\n$/ && $paren == 0)
		{
			return Config4GNU::DNS::ZoneFileParser::Entry->new(\@entry_tokens);
		}
	}
	
	#end of file reached
	if (@entry_tokens)
	{
		return Config4GNU::DNS::ZoneFileParser::Entry->new(\@entry_tokens);
	}
	else
	{
		return undef;
	}
}

=item my $filename = $self->filename

Returns the name of the file being parsed.

=cut

sub filename
{
	my $self = shift;
	return $self->{'FILENAME'};
}

=back

=cut

package Config4GNU::DNS::ZoneFileParser::Entry;

=head2 Config4GNU::DNS::ZoneFileParser::Entry

=head3 Methods

=over

=cut

sub new
{
	my $class = shift;
	my $self = { 'TOKENS' => shift };
	bless $self, $class;
}

=item $entry->isComment

Returns nonzero if the entry only contains a comment and/or whitespace.

=cut

sub isComment
{
	my $self = shift;
	foreach my $token (@{$self->tokens})
	{
		return 0 if ($token !~ /^\s/ && $token !~ /^;/);
	}
	return 1;
}

=item $entry->toString

Returns the content as found in the original file of the entry.

=cut

sub toString
{
	my $self = shift;
	return join('', @{$self->tokens});
}

=item my $token = $entry->token($n);

Accesses the "token" at position $n, where "token" is any part of the string
that is not broken down into a smaller component at the parser level. This
includes words, whitespace, comments, and grouping operators. If the
index given is invalid, the result is C<undef>.

=cut

sub token
{
	my $self = shift;
	my $index = shift;
	return $self->tokens->[$index];
}

# returns an array reference
sub tokens
{
	my $self = shift;
	return $self->{'TOKENS'};
}

=item my $word = $entry->word($n);

Accesses the "word" at position $n, where "word" is a character string that
has meaning to the computer (i.e. not whitespace, a comment, or grouping
operator). If the index given is invalid, the result is C<undef>.

=cut

sub word
{
	my $self = shift;
	my $index = shift;
	my @words = $self->words;
	return $words[$index];
}

=item my @words = $entry->words;

Returns an array composed of the "words" that make up the current entry.

=cut

sub words
{
	my $self = shift;
	if (defined $self->{'WORDS'})
	{
		return @{$self->{'WORDS'}};
	}
	my @words = ();
	foreach my $word (@{$self->tokens})
	{
		if ($word =~ /^\s/ || $word =~ /^;/)
		{
			# whitespace or comment
			next;
		}
		if ($word eq '(' || $word eq ')')
		{
			# ignore parenthesis
			next;
		}
		push @words, $word;
	}
	$self->{'WORDS'} = \@words;
	return @words;
}

=back

=head2 Config4GNU::DNS::ZoneFileParser::EntryBase

=head3 Methods

=over

=cut

package Config4GNU::DNS::ZoneFileParser::EntryBase;

sub new
{
	my $class = shift;
	my $tokenlist = shift;
	my $self = { 'TOKENS' => $tokenlist };
	return bless $self, $class;
}

sub setComment
{
	my $self = shift;
	my $comment = shift;
	$self->{'COMMENT'} = $comment;
}

=item $entry->data

Returns the data portion of the entry.

=cut

sub data
{
	my $self = shift;
	return join('', @{$self->{'TOKENS'}->tokens});
}

sub toString
{
	my $self = shift;
	my $comment = defined($self->{'COMMENT'}) ? $self->{'COMMENT'} : '';
	return $comment . $self->data;
}

=back

=head2 Config4GNU::DNS::ZoneFileParser::CommandEntry

Extends Config4GNU::DNS::ZoneFileParser::EntryBase.

=cut

package Config4GNU::DNS::ZoneFileParser::CommandEntry;
use base 'Config4GNU::DNS::ZoneFileParser::EntryBase';

sub create
{
	my $tokenlist = shift;
	my $firsttoken = $tokenlist->token(0);
	if ($firsttoken !~ /^\$/)
	{
		warn "Not a command entry";
		return undef;
	}
	
	my $self = { 'TOKENS' => $tokenlist };
	if ($firsttoken eq '$TTL')
	{
		return bless $self, 'Config4GNU::DNS::ZoneFileParser::TtlCommandEntry';
	}
	else
	{
		return bless $self, 'Config4GNU::DNS::ZoneFileParser::CommandEntry';
	}
}

package Config4GNU::DNS::ZoneFileParser::TtlCommandEntry;
use base 'Config4GNU::DNS::ZoneFileParser::CommandEntry';

sub getTtl
{
	my $self = shift;
	return $self->{'TOKENS'}->word(1);
}

=head2 Config4GNU::ZoneFileParser::ResourceRecord

Extends Config4GNU::DNS::ZoneFileParser::EntryBase.

=head3 Methods

=over

=cut

package Config4GNU::DNS::ZoneFileParser::ResourceRecord;
use base 'Config4GNU::DNS::ZoneFileParser::EntryBase';

=item $record->setCurrentDomainName('www');

Sets the current domain name. This domain name is used if the parsed entry
does not specify a domain name.

=cut

sub setCurrentDomainName
{
	my $self = shift;
	$self->{'currentDomainName'} = shift;
}

=item $record->setOrigin('example.org.');

This specifies the origin of the current domain name. This value is used
to determine the fully-qualified domain name of the owner domain and domain
names specified in the record data if they are not fully-qualified themselves.

=cut

sub setOrigin
{
	my $self = shift;
	$self->{'ORIGIN'} = shift;
}

=item $record->setCurrentTtl('1d');

Sets the current TTL. This TTL is used if the parsed entry does not specify
a TTL.

=cut

sub setCurrentTtl
{
	my $self = shift;
	$self->{'currentTtl'} = shift;
}

=item $record->setCurrentClass('IN');

Sets the current class (e.g. 'IN'). This class is used if the parsed entry
does not specify a class.

=cut

sub setCurrentClass
{
	my $self = shift;
	my $class = shift;
	$self->{'currentClass'} = $class;
	warn "Unrecognized class $class"
		unless ($self->isClass($class));
}

=item my $owner = $record->domainName;

Returns the fully-qualified domain name that owns this resource record.

=cut

sub domainName
{
	my $self = shift;
	my $firstToken = $self->{'TOKENS'}->tokens->[0];
	if ($firstToken =~ /^\s/)
	{
		return $self->{'currentDomainName'};
	}
	else
	{
		if ($firstToken eq '@')
		{
			return $self->{'ORIGIN'};
		}
		elsif ($firstToken =~ /\.$/)
		{
			return $firstToken;
		}
		else
		{
			return $firstToken . '.' . $self->{'ORIGIN'};
		}
	}
}

=item my $class = $record->class;

Returns the record class, e.g. 'IN'.

=cut

sub class
{
	my $self = shift;
	my ($ttl, $class, $type, $rdata) = $self->getParts;
	return $class;
}

=item my $type = $record->type;

Returns the record type, e.g. 'A' or 'PTR'.

=cut

sub type
{
	my $self = shift;
	my ($ttl, $class, $type, $rdata) = $self->getParts;
	return $type;
}

=item my $ttl = $record->ttl;

Returns the record's effective TTL as a numeric value. The record's TTL
is expressed in number of seconds.

=cut

sub ttl
{
	my $self = shift;
	my ($ttl, $class, $type, $rdata) = $self->getParts;
	return parseduration($ttl);
}

sub parseduration
{
	my $duration = shift;

	my $SECONDS_PER_MINUTE = 60;
	my $SECONDS_PER_HOUR = 60 * $SECONDS_PER_MINUTE;
	my $SECONDS_PER_DAY = 24 * $SECONDS_PER_HOUR;
	my $SECONDS_PER_WEEK = 7 * $SECONDS_PER_DAY;

	my $seconds = 0;
	if ($duration =~ /^(\d+)w/)
	{
		$seconds += ($1 * $SECONDS_PER_WEEK);
		$duration = $';
	}
	if ($duration =~ /^(\d+)d/)
	{
		$seconds += ($1 * $SECONDS_PER_DAY);
		$duration = $';
	}
	if ($duration =~ /^(\d+)h/)
	{
		$seconds += ($1 * $SECONDS_PER_HOUR);
		$duration = $';
	}
	if ($duration =~ /^(\d+)m/)
	{
		$seconds += ($1 * $SECONDS_PER_MINUTE);
		$duration = $';
	}
	if ($duration =~ /^(\d+)s?$/)
	{
		$seconds += $1;
	}
	elsif ($duration ne '')
	{
		warn "unrecognized TTL $duration\n";
	}

	return $seconds;
}

=item my ($ttl, $class, $type, $rdata) = $record->getParts;

Returns four parsed values from the resource record. The first three are
simple strings. For each one, if the value was specified, the value is returned
exactly as it appeared in the config file (e.g. a TTL of '1d' is returned as
'1d'). If the value is not specified, the effective value is returned.
The final value, $rdata, is an array reference containing the remaining
tokens in the entry.

=cut

sub getParts
{
	my $self = shift;

	my $ttl = $self->{'currentTtl'};
	my $class = $self->{'currentClass'};

	my @tokens = @{$self->{'TOKENS'}->tokens};
	
	# skip past domain name, if any
	$self->nextword(\@tokens);

	foreach (0, 1)
	{
		# check for TTL
		if ($tokens[0] =~ /^\d/)
		{
			# starts with a digit, so it must be TTL
			$ttl = $tokens[0];
			$self->nextword(\@tokens);
		}
		# check for valid class
		elsif ($self->isClass($tokens[0]))
		{
			$class = $tokens[0];
			$self->nextword(\@tokens);
		}
		else
		{
			last;
		}
	}

	# next token must be type
	my $type = $tokens[0];
	$self->nextword(\@tokens);

	return ($ttl, $class, $type, \@tokens);
}

=item $text = $record->rr;

Returns the resource record minus the owner domain name as it appears in the
config file.

=cut

sub rr
{
	my $self = shift;
	# rr starts after the domain name
	my @tokens = @{$self->{'TOKENS'}->tokens};
	$self->nextword(\@tokens);
	return join('', @tokens);
}

sub rdataAsString
{
	my $self = shift;
	return join(' ', $self->rdataAsWords);
}

sub rdataAsWords
{
	my $self = shift;
	my ($ttl, $class, $type, $rdata) = $self->getParts;
	my @words = ();
	foreach my $word (@{$rdata})
	{
		if ($word =~ /^\s/ || $word =~ /^;/)
		{
			# whitespace or comment
			next;
		}
		if ($word eq '(' || $word eq ')')
		{
			# ignore parenthesis
			next;
		}
		push @words, $word;
	}
	return @words;
}

# internal function
# returns nonzero if the given string is a valid class identifier
sub isClass
{
	my $self = shift;
	my $potentialClass = shift;

	return ($potentialClass eq 'IN');
}

# internal function
sub nextword
{
	my $self = shift;
	my $tokens = shift;
	while ($tokens->[0] !~ /^\s/)
	{
		shift @{$tokens};
	}
	while ($tokens->[0] =~ /^\s/)
	{
		shift @{$tokens};
	}
	return $tokens;
}

=back

=head1 BUGS

Many command entries are not supported, although they should parse and resave
without difficulty. Examples include $ORIGIN and $INCLUDE.

Backslashes to escape characters are only supported inside quoted character
strings.

The C<\DDD> format of character escapes are not supported anywhere.

=head1 AUTHOR

Jason Long <jason@long.name>

=cut

1;
