#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Config4GNU::AbstractNode;

=head1 NAME

Config4GNU::AbstractNode - a base class for Config4GNU provider nodes.

=head1 DESCRIPTION

All provider nodes extend this class, either directly or through another
class that does. If you want to write a parser for a brand new configuration
file format, extend this class. Create the necessary methods to read and
write the configuration files and override some of the methods if
necessary.

If you need a parser that behaves similarly to an existing parser, try to
find a way to extend that parser, so that you do not need to duplicate code.

=head2 METHODS

=over 4

=cut

# create a new instance with the specified filename
sub new
{
    my ($class, $file) = @_;
    my $self = {};
    $self->{FILENAME} = $file;
    bless $self, $class;

	$self->init;
	return $self;
}

sub init
{
    my $self = shift;

    $self->{CHILDREN} = [];
}

=item $node->children

Returns an array reference containing the child nodes of this node.
Each child node is of type CFG::AbstractNode, and is guaranteed to have
a unique name to this container.

=cut

sub children
{
    my $self = shift;

	my @result = ();
	if (defined $self->{CHILDREN})
	{
		foreach my $i (@{$self->{CHILDREN}})
		{
			if ($i->isa("Config4GNU::AbstractNode"))
			{
				push @result, $i;
			}
		}
	}
	return \@result;
}

=item $node->properties

Returns an array reference containing the properties of this node.
Each property is of type CFG::AbstractProperty. Properties have "IDs"
that are unique to all the properties of this node.

=cut

sub properties
{
    my $self = shift;

	my @result = ();
	if (defined $self->{CHILDREN})
	{
		foreach my $i (@{$self->{CHILDREN}})
		{
			if ($i->isa("Config4GNU::AbstractProperty"))
			{
				push @result, $i;
			}
		}
	}
	return \@result;
}

=item $node->allowedChildTypes

Returns information describing what types of child objects are allowed
in this node. The nature of this information has not been defined yet.
The default implementation returns an empty array, which means users
cannot create new objects in this node.

=cut

sub allowedChildTypes
{
	my $self = shift;
	
	my @empty = ();
	return \@empty;
}

=item $node->allowedProperties

Returns information describing what properties are allowed in this
node. The nature of this information has not been defined yet.
The default implementation returns an empty array, which means users
cannot add properties to this node.

=cut

sub allowedProperties
{
	my $self = shift;
	
	my @empty = ();
	return \@empty;
}

=item $node->name

Returns the name of this node. By default, this is the value of
$node->{NAME}. Subclassers are encouraged to override this method
if you want different behavior. The node's name is guaranteed to be
unique to all the children of the parent node.

=cut

sub name
{
    my $self = shift;
    return $self->{NAME};
}

=item $node->setName($newName)

Sets the name of this node. If the node has a parent node, then its name
must be unique to all the children of the parent.

=cut

sub setName
{
	my $self = shift;
	my $newname = shift;

	$self->{NAME} = $newname;
}

=item $node->id

Returns the ID of the node. The ID is the parent's ID concatenated with this
node's name.

=cut

sub id
{
	my $self = shift;
	
	return '/' unless (defined $self->parent);
	my $parentId = $self->parent->id;
	$parentId .= '/' unless ($parentId =~ /\/$/);
	return $parentId . $self->name;
}

=item $node->find

Given an ID, return the node or property that has that ID.

=cut

sub find
{
	my $self = shift;
	my $id = shift;
	
	if ($id =~ /^\//)
	{
		# id is absolute
		if (defined($self->parent))
		{
			warn "find called with absolute path on nonroot node\n";
		}
		# call myself recursively with the remainder of id
		return $self->find($');
	}
	if ($id =~ /^([^\/]+)\//)
	{
		# look for child node
		my $nextPart = $1;
		my $nextId = $';
		
		my @children = @{$self->children};
		foreach my $child (@children)
		{
			if ($child->name eq $nextPart)
			{
				return $child->find($nextId);
			}
		}
		# not found
		return undef;
	}
	else
	{
		# look for property
		my @properties = @{$self->properties};
		foreach my $prop (@properties)
		{
			if ($prop->name eq $id)
			{
				return $prop;
			}
		}
		# not found
		return undef;
	}
}

=item $node->parent

Returns the parent node.

=cut

sub parent
{
	my $self = shift;
	return $self->{PARENT};
}

=item $node->setParent($parent)

Sets this node's parent.

=cut

sub setParent
{
	my $self = shift;
	my $parent = shift;
	
	$self->{PARENT} = $parent;
}

=item $node->addChildObject($childObject)

Given a child that was already created, using an appropriate create
method, add it to this node's list of children.

=cut

sub addChildObject
{
	my $self = shift;
	my $child = shift;
	
	# add it to the CHILDREN array
	push @{$self->{CHILDREN}}, $child;
	
	$child->setParent($self);
}

=item $node->removeChild($child)

Remove the specified child object from this node.

=cut

sub removeChild
{
	my $self = shift;
	my $childToDelete = shift;

	my $len = @{$self->{CHILDREN}};
	for (my $i = 0; $i < $len; $i++)
	{
		if ($self->{CHILDREN}[$i] == $childToDelete)
		{
			splice(@{$self->{CHILDREN}}, $i, 1);
			return;
		}
	}
			
	warn "removeChild: could not find specified node!\n";
}

=item $node->addPropertyObject($propertyObject)

Given a property that was already created, using an appropriate create
method, add it to this node's list of properties.

=cut

sub addPropertyObject
{
    my $self = shift;
	$self->addChildObject(@_);
}

=item $node->removeProperty($property)

Remove the specified property object from this node.

=cut

sub removeProperty
{
	my $self = shift;
	$self->removeChild(@_);
}

=item $node->getProperty($propertyname)

Returns the specified property object.

=cut

sub getProperty
{
	my $self = shift;
	my $propertyname = shift;

	foreach my $prop (@{$self->properties})
	{
		return $prop if ($prop->name eq $propertyname);
	}
	return undef;
}

=item $value = $node->getPropertyValue($propertyname)

Retrieves the defined property's value and returns it, or C<undef> if the
property is not defined.

=cut

sub getPropertyValue
{
	my $self = shift;
	my $propertyname = shift;

	my $prop = $self->getProperty($propertyname);
	if (defined $prop)
	{
		return $prop->value;
	}
	return undef;
}

=back

=cut

# return true value
1;
