#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Config4GNU::AbstractProperty;

=head1 NAME

Config4GNU::AbstractProperty - a base class for Config4GNU properties.

=head1 DESCRIPTION

All property implementations extend this class, either directly or
through another class that does. When you implement a new type of
property, you'll want to extend this class. Otherwise, use an existing
class.

=head2 CONSTRUCTOR

=over 4

=item $property = Config4GNU::Property->new($name, $value);

Creates a new property and assigns it a name and value.

=cut

sub new
{
    my $class = shift;
		
	# arguments
    my ($prop_name, $prop_value) = @_;

    my $self = {
        NAME => $prop_name,
		VALUE => $prop_value };
	
    return bless $self, $class;
}

=back

=head2 METHODS

=over 4

=item $property->name

Returns the name of this property. By default, this is the value of
$property->{NAME}. Subclassers are encouraged to override this method
if you want different behavior.

The name of the property can be undefined; this is called an
"anonymous property". Use this if a node contains only a list of items,
and property "names" do not apply.

=cut

sub name
{
    my $self = shift;
    return $self->{NAME};
}

=item $property->setName($newName)

Renames this property. The name cannot contain any of these characters: =[]

=cut

sub setName
{
	my $self = shift;
	$self->{NAME} = shift;
}

=item $property->value

Returns a stringified representation of this property's value.
The meaning of a property's "value" is dependent on the property type.
The default behavior simply stores a string value.

=cut

sub value
{
    my $self = shift;
    return $self->{VALUE};
}

=item $property->setValue($newValue)

Sets a new value for the property. The input is a string value; it should
be converted to a different data type if necessary. Subclassers can
override this method and may also provide additional methods to set
the property's value, e.g. setBooleanValue, setNumericValue.

=cut

sub setValue
{
	my $self = shift;
	my $newvalue = shift;
	$self->{VALUE} = $newvalue;
}

=item $node->id

Returns the ID of the property. The ID is the parent's ID concatenated with this
property's name.

=cut

sub id
{
	my $self = shift;
	
	return (defined($self->parent) ? $self->parent->id . '/' : '/')
		. $self->name;
}

=item $node->parent

Returns the parent node.

=cut

sub parent
{
	my $self = shift;
	return $self->{PARENT};
}

=item $node->setParent($parent)

Sets this node's parent.

=cut

sub setParent
{
	my $self = shift;
	my $parent = shift;
	
	$self->{PARENT} = $parent;
}

=item $property->writeConfig($fh)

Outputs this property to the given file handle in the configuration file's
native format. The default implementation of this method does nothing.

=cut

sub writeConfig
{
    my $self = shift;
	my $fh = shift;

	# output this property to the given file handle $fh
}

=back

=cut

# return true value
1;
