#!/usr/bin/perl -w

use strict;
use warnings;

use Config4GNU::AbstractProvider;
use Config4GNU::AbstractProperty;
use Config4GNU::FileIOException;

package Config4GNU::ISC::Section;
use base 'Config4GNU::AbstractNode';

# create a new instance
sub new
{
    my $class = shift;
		
    my $self = {
        'WORDS' => \@_ };
    return bless $self, $class;
}

sub createSection
{
	my $self = shift;
    return Config4GNU::ISC::Section->new(@_);
}

sub createProperty
{
	my $self = shift;
    return Config4GNU::ISC::Property->new(@_);
}

sub name
{
	my $self = shift;
    return join(' ', $self->words);
}

sub words
{
	my $self = shift;
	my @wordsmodified = map { Config4GNU::ISC::Word::unquote($_) }
			@{$self->{WORDS}};
	return @wordsmodified;
}

sub word
{
	my $self = shift;
	my $i = shift;
	my @words = $self->words;
	return $words[$i];
}

sub writeConfig
{
	my $self = shift;
	my $fh = shift;

	# for each word in the statement opening up this section, print out
	# the preceding comments/whitespace and then print the actual word
	my $numWords = @{$self->{WORDS}};
	for (my $i = 0; $i < $numWords; $i++)
	{
		print $fh @{$self->{COMMENTS}}[$i];
		print $fh @{$self->{WORDS}}[$i];
	}
	
	# print whitespace before brace and actual open-brace character
	print $fh @{$self->{COMMENTS}}[$numWords], '{';

	# now iterate through each child of this section (which could be statements
	# or subsections)
    my @children = @{$self->{CHILDREN}};
    foreach my $child (@children)
    {
        $child->writeConfig($fh);
    }

	# print section closing
    print $fh $self->{END_COMMENTS};	
	print $fh "}";
}

package Config4GNU::ISC;
use base ('Config4GNU::ISC::Section', 'Config4GNU::AbstractProvider');

sub new
{
    my ($class, $file) = @_;
    my $self = {};
    $self->{FILENAME} = $file;
    bless $self, $class;

	$self->init;
	return $self;
}

sub init
{
    my $self = shift;

    my $parser = Config4GNU::ISC::Parser->new($self->{FILENAME});
	
	# the configuration file can have nested sections... this array keeps
	# track of which sections we're currently in
	my @sectionStack = ();
	my $curSection = $self;
	
	# these variables store words and comments/whitespace as they are read
	# until the object to manage them is created
	my $curWord = 0;
	my @words = ();
	my @comments = ();

	while ($parser->hasMoreTokens)
	{
		my $token = $parser->nextToken;
		if ($parser->isComment($token) || $parser->isWhitespace($token))
		{
			# since comments can occur in between words, store them in an
			# array at the same index number that the next word will be
			# stored.
			$comments[$curWord] .= $token;
		}
		elsif ($token eq '{')
		{
			# begin a new section

            # create the backend object
			my $servant = $curSection->createSection(@words);
			my @commentscopy = @comments;
    		$servant->{COMMENTS} = \@commentscopy;

			# reset comments & tokens
			@comments = ();
			@words = ();
			$curWord = 0;
			
			# add it to current section's children
			$curSection->addChildObject($servant);
			
			push @sectionStack, $curSection;
			$curSection = $servant;
		}
		elsif ($token eq '}')
		{
			# end current section
			# FIXME... handle the case that values are in @words
			warn "statement at end of section\n" if (@words);
			$curSection->{END_COMMENTS} = join('', @comments);
			
			# reset comments and words
			@comments = ();
			@words = ();
			$curWord = 0;
			
			# pull the last section off the stack
			$curSection = pop @sectionStack;
			
			# if we pulled them all off, revert to using $self as the current
			# section
			$curSection = $self unless (defined $curSection);
		}
		elsif ($token eq ';')
		{
			# end current statement
			if (@words)
			{
	            # create the backend object for property
				my $servant = $curSection->createProperty(@words);
				my @commentscopy = @comments;
    			$servant->{COMMENTS} = \@commentscopy;

				# reset comments & tokens
				@comments = ();
				@words = ();
				$curWord = 0;
				
            	# add it to the current section's children
				$curSection->addChildObject($servant);
			}
			else
			{
				# found a semicolon, but no words have been read... so instead,
				# just treat it as whitespace or a comment
				$comments[$curWord] .= ';';
			}
		}
		else
		{
			# found a word...
			$words[$curWord] = $token;
			$curWord++;
			$comments[$curWord] = '';
		}
	}

	# FIXME... any values in @words will be lost here
    $self->{END_COMMENTS} = join('', @comments);
}

sub save
{
    my $self = shift;

	open OUTFILE, '>', $self->{FILENAME} . ".tmp"
			or throw Config4GNU::FileIOException errorMessage => "$!";
	$self->writeConfig(\*OUTFILE);
	close OUTFILE;
}

sub writeConfig
{
	my $self = shift;
	my $fh = shift;
	
    my @children = @{$self->{CHILDREN}};
    for my $prop (@children)
    {
        $prop->writeConfig($fh);
    }
    print $fh $self->{END_COMMENTS};
}

sub name
{
	return "Root";
}

package Config4GNU::ISC::Parser;

=over

=cut

sub new
{
	my $class = shift;
	my $filename = shift;
	
	my $self = {};
	local $/;
	open INFILE, '<', $filename
            or throw Config4GNU::FileIOException errorMessage => "$!";
	$self->{TOKENS} = [];
	$self->{BUFFER} = <INFILE>;
	close INFILE;
	
	bless $self, $class;
	return $self;
}

sub hasMoreTokens
{
	my $self = shift;
	my @next_tokens = @{$self->{TOKENS}};
	return scalar(@next_tokens) > 0 || length($self->{BUFFER}) > 0;
}

=item my $token = $parser->nextToken;

Grab the next interesting thing from the config file and return it.

=cut

sub nextToken
{
	my $self = shift;
	if (@{$self->{TOKENS}})
	{
		return shift @{$self->{TOKENS}};
	}
	return undef unless (length($self->{BUFFER}));

	my $buf = $self->{BUFFER};
	my $c = substr($buf, 0, 1);
	if ($c eq ' ' || $c eq "\t" || $c eq "\n")
	{
		# this token is whitespace... continue grabbing characters until
		# I get a non-whitespace character
		$buf =~ /^\s+/s;
		$self->{BUFFER} = substr($buf, length($&));
		return $&;
	}
	if (substr($buf, 0, 2) eq '/*')
	{
		# multi-line comment   (i.e. /* ... */ )
		$buf =~ /^\/\*.*?\*\/\s*/s;
		$self->{BUFFER} = substr($buf, length($&));
		return $&;
	}
	if (substr($buf, 0, 2) eq '//')
	{
		# single-line comment (C++ style)
		$buf =~ /^\/\/.*?\n\s*/s;
		$self->{BUFFER} = substr($buf, length($&));
		if (defined $2)
		{
			unshift @{$self->{TOKENS}}, $2;
		}
		return $&;
	}
	if ($c eq '#')
	{
		# single-line comment (shell style)
		$buf =~ /^#.*?\n\s*/s;
		$self->{BUFFER} = substr($buf, length($&));
		return $&;
	}
	if (index(';{},', $c) != -1)
	{
		# single-character token
		$self->{BUFFER} = substr($buf, 1);
		return $c;
	}
	if ($c eq '"')
	{
		# quoted string
		$buf =~ /^"[^\\"]*"?/s;		## FIXME: won't accept "sometext\\"
		$self->{BUFFER} = substr($buf, length($&));
		return $&;
	}
	if ($buf =~ /^([\w.:*-]+)(\s*)/s)
	{
		# a word
		$self->{BUFFER} = substr($buf, length($&));
		if (defined $2)
		{
			unshift @{$self->{TOKENS}}, $2;
		}
		return $1;
	}
	
	# um... getting down here means there's something in the configuration
	# file that is an unrecognized syntax
	die "unparseable? $buf\n";
	return $buf;
}

=item $parser->isComment($token);

Returns nonzero if the token $token is considered a comment.

=cut

sub isComment
{
	my $self = shift;
	my $token = shift;
	
	return ($token =~ /^\/\// || $token =~ /^\/\*/ || $token =~ /^#/);
}

=item $parser->isWhitespace($token);

Returns nonzero if the token $token is considered whitespace.

=cut

sub isWhitespace
{
	my $self = shift;
	my $token = shift;
	
	return ($token =~ /^\s/);
}

=back

=cut

package Config4GNU::ISC::Property;
use base 'Config4GNU::AbstractProperty';

# create a new instance
sub new
{
    my $class = shift;
		
    my $self = {
        'WORDS' => \@_,
		'IS_PROPERTY' => 1 };
	if (@_ == 1)
	{
		$_[1] = $_[0];
		$_[0] = undef;
		$self->{'HAS_NAME'} = 0;
	}
	else
	{
		$self->{'HAS_NAME'} = 1;
	}
    return bless $self, $class;
}

sub name
{
    my $self = shift;
    return @{$self->{WORDS}}[0];
}

sub value
{
    my $self = shift;
	my @words = @{$self->{WORDS}};
	shift @words;
	my @wordsmodified = map { Config4GNU::ISC::Word::unquote($_) }
			@words;
    return join(' ', @wordsmodified);
}

sub setValue
{
	my $self = shift;
	my $newvalue = shift;
	$self->{value} = $newvalue;
}

sub writeConfig
{
    my $self = shift;
	my $fh = shift;
		
	# for each word in the statement, print out
	# the preceding comments/whitespace and then print the actual word
	my $numWords = @{$self->{WORDS}};
	for (my $i = 0; $i < $numWords; $i++)
	{
		if (defined $self->{COMMENTS}->[$i])
		{
			print $fh $self->{COMMENTS}->[$i];
		}
		if (defined($self->{WORDS}->[$i]))
		{
			print $fh $self->{WORDS}->[$i];
		}
	}
	
	if (defined($self->{COMMENTS}->[$numWords]))
	{
		print $fh $self->{COMMENTS}->[$numWords];
	}
	print $fh ";";
}

package Config4GNU::ISC::Word;
# this is not really a class/object at this time, rather a place to put
# functions dealing with individual words

sub isQuoted
{
	my $word = shift;
	
	return ($word =~ /^"/);
}

sub quote
{
	my $word = shift;
	
	$word =~ s/(["\\])/\\$1/g;
	return "\"$word\"";
}

sub unquote
{
	my $word = shift;
	
	return $word unless (isQuoted $word);
	$word =~ s/\\(.)/$1/g;
	return substr($word, 1, -1);
}

1;

=head1 NAME

Config4GNU::ISC - read/write files in the format of ISC software config files.

=head1 SYNOPSIS

  use Config4GNU::ISC;
  my $config = Config4GNU::ISC->new('/etc/named.conf');

  # change a setting
  my $options_section = $config->getSection('options');
  my $oldvalue = $options_section->getPropertyValue('recursion');
  $options_section->setPropertyValue('recursion', 'no');

  # add a new section
  my $zone = $config->newSection('zone', 'example.com');
  $zone->setPropertyValue('type', 'master');
  $zone->setPropertyValue('file', 'db.example.com');

  # delete a section
  $config->deleteSection('zone', 'example.org');

  # write the changes back to the file
  $config->save;

=head1 DESCRIPTION

This module provides parsing ability for configuration files that use the
Internet Software Consortium configuration file syntax. For example, BIND
version 8 or higher and the ISC DHCP server use this syntax.

=head1 HACKING

Config4GNU::ISC::init is where the parsing takes place, but you may
also want to look at Config4GNU::ISC::Parser which does the actual
reading from the file and returns small, manageable chunks called "tokens."

Config4GNU::ISC::Section is the implementation that manages each
section found in the file (a section begins with one or more words and an open
brace and ends with a closing brace and has a number of statements or child
sections in between).

Config4GNU::ISC::Statement is the implementation that manages the
statements found within the sections. Each statement is composed of one or more
words and ends with a semicolon.

=head1 SEE ALSO

See the Config4GNU website at http://config4gnu.sourceforge.net/

=head1 AUTHOR

Jason Long <jason@long.name>

=cut
