#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use Config4GNU::AbstractProvider;
use Config4GNU::AbstractProperty;
use Config4GNU::FileIOException;
use strict;
use warnings;

package Config4GNU::Ini;
use base 'Config4GNU::AbstractProvider';

sub init
{
    my $self = shift;
	$self->SUPER::init;

    open INIFILE, '<', $self->{FILENAME}
            or throw Config4GNU::FileIOException errorMessage => "$!";
    my $section;
    my $comments = '';
    while (my $line = <INIFILE>)
    {
		chomp $line;
		
		# TODO: handle line extenders (backslash at the end of the line)
		
        # check for a comment
        if ($line =~ /^\s*[#;]/)
        {
            # ignore the comment
            $comments .= "$line\n";
        }
        # check for section header
        elsif ($line =~ /^(\s*)\[([^\]]*)\](\s*)$/)
        {
            my $whitespace1 = $1;
			my $section_name = $2;
			my $whitespace2 = $3;
			
            # create the backend object
			$section = $self->createSection($section_name);
    		$section->{COMMENTS} = $comments;
			$section->{whitespace1} = $whitespace1;
			$section->{whitespace2} = $whitespace2;

			# reset comments
			$comments = '';

            # add the section
			$self->addChildObject($section);
        }
        # check for property assignment
        # FIXME: the following regular expression does not handle when
        # the property name has internal spaces
        elsif ($line =~ /^(\s*)([ \w.]*?)(\s*=\s*)(.*)$/ && defined($section))
        {
            my $left_whitespace = $1;
			my $prop_name = $2;
			my $equalspart = $3;
            my $prop_value = $4;

			# use the current section to create the property implementation
			my $prop = $section->createProperty($prop_name, $prop_value);
			$prop->{comments} = $comments;
			$prop->{whitespace} = $left_whitespace;
			$prop->{equalspart} = $equalspart;
			
			# reset comments
            $comments = '';
			
			# add the property to the current section
			$section->addPropertyObject($prop);
        }
        else
        {
            $comments .= "$line\n";
        }
    }
    close INIFILE;

    $self->{END_COMMENTS} = $comments;
}

=item $ini->save

Writes the configuration file.

=cut

sub save
{
    my $self = shift;

	open OUTFILE, '>', $self->{FILENAME}
			or throw Config4GNU::FileIOException errorMessage => "$!";

    my @sections = @{$self->children};
    for my $section (@sections)
    {
        $section->writeConfig(\*OUTFILE);
    }
    print OUTFILE $self->{END_COMMENTS};
	
	close OUTFILE;
}

# extend this method...
sub createSection
{
	my $self = shift;
	my ($section_name) = @_;

    return Config4GNU::Ini::Section->new($section_name);
}

package Config4GNU::Ini::Section;
use base 'Config4GNU::AbstractNode';

# create a new instance with the specified name
sub new
{
    my $class = shift;
    my $self = {};
		
	# arguments
	my ($name) = @_;

    $self->{NAME} = $name;
	bless $self, $class;
	
	$self->init;
    return $self;
}

sub writeConfig
{
    my $self = shift;
	my $fh = shift;
	
    # sanity check parts of section
	unless (defined $self->{COMMENTS})
	{
		$self->{COMMENTS} = "\n";
	}
	
	unless (defined $self->{whitespace1})
	{
		$self->{whitespace1} = '';
	}
	unless (defined $self->{whitespace2})
	{
		$self->{whitespace2} = '';
	}

	# output section comments and header
	print $fh $self->{COMMENTS};
    print $fh $self->{whitespace1}, "[", $self->name, "]",
			$self->{whitespace2}, "\n";

	# iterate through each property in this section
    foreach my $prop (@{$self->properties})
    {
		$prop->writeConfig($fh);
    }
}

sub setProperty
{
	my $self = shift;
	my ($name, $value) = @_;
	
	foreach my $prop (@{$self->properties})
	{
		if ($prop->{NAME} eq $name)
		{
			$prop->{VALUE} = $value;
			return;
		}
	}
	
	# add the property to the end
	$self->addProperty($name, $value);
}

sub addProperty
{
	my $self = shift;
	my $propertyUid = shift;
	
	my $property = $self->createProperty($propertyUid);
	$self->addPropertyObject($property);
	return $property;
}

# extend this method...
sub createProperty
{
	my $self = shift;
	my ($prop_name, $prop_value) = @_;
	
	return Ini::Property->new($prop_name, $prop_value);
}

package Config4GNU::Ini::Property;
use base 'Config4GNU::AbstractProperty';

sub writeConfig
{
    my $self = shift;
	my $fh = shift;
		
	# sanity check parts of each property
	# TODO: check comments
	#		for each line, should begin with valid comment character
	#		should end with a linefeed
	if (!defined($self->{comments}))
	{
		$self->{comments} = '';
	}
	if (!defined($self->{whitespace}) || $self->{whitespace} =~ /\S/)
	{
		# whitespace contains a nonwhitespace character, so reset it
		$self->{whitespace} = '';
	}
	if (!defined($self->{equalspart}) || $self->{equalspart} !~ /^\s*=\s*$/)
	{
		# invalid comment string, so reset it
		$self->{equalspart} = ' = ';
	}

	# output the property
    print $fh $self->{comments};
    print $fh $self->{whitespace}, $self->{NAME}, $self->{equalspart},
			$self->{VALUE}, "\n";
}

1;
