#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use Config4GNU::SimpleKeyedValues;
use Config4GNU::Nagios::Config;
use Config4GNU::Nagios::Status;
use strict;

package Config4GNU::Nagios;
use base 'Config4GNU::AbstractProvider';
$Config4GNU::Nagios::DEFAULT_FILE = '/usr/local/nagios/etc/nagios.cfg';

# create a new instance with the specified filename
sub new
{
    my ($class, $file) = @_;
	unless (defined $file)
	{
		$file = $Config4GNU::Nagios::DEFAULT_FILE;
	}
	return $class->SUPER::new($file);
}

sub init
{
	my $self = shift;
	
	my $mainconfig = Config4GNU::Nagios::Config::MainConfig->new(
			$self->{FILENAME});
	$mainconfig->{NAME} = 'Main Configuration';
	$self->addChildObject($mainconfig);
	
	foreach my $cfgfile_directive (@{$mainconfig->properties})
	{
		my $directive_name = $cfgfile_directive->name;
		my $directive_value = $cfgfile_directive->value;
		if ($directive_name eq 'status_file')
		{
			$self->{STATUSFILENAME} = $directive_value;
		}
		elsif ($directive_name eq 'cfg_file')
		{
			my $objectconfig =
				Config4GNU::Nagios::Config::ObjectConfig->new($directive_value);
			$directive_value =~ /([^\/]*)$/;
			$objectconfig->{NAME} = $1;
			$self->addChildObject($objectconfig);
			
			$self->importObjectConfiguration($objectconfig);
		}
	}
	
	if (defined $self->{STATUSFILENAME})
	{
		$self->{STATUSFILE} = Config4GNU::Nagios::Status->new(
				$self->{STATUSFILENAME});
		$self->importStatusFile($self->{STATUSFILE});
	}
	else
	{
		warn "status_file not defined\n";
	}
}

sub importObjectConfiguration
{
	my $self = shift;
	my $objectconfig = shift;
	
	foreach my $section (@{$objectconfig->children})
	{
		if ($section->{TYPE} eq 'hostgroup')
		{
			my $servant = Config4GNU::Nagios::HostGroup->new($section);
			push @{$self->{hostgroups}}, $servant;
			$servant->{MAIN} = $self; #creates a circular reference (I think)
		}
		elsif ($section->{TYPE} eq 'host' && !$section->isTemplate)
		{
			my $servant = Config4GNU::Nagios::Host->new($section);
			push @{$self->{hosts}}, $servant;
		}
	}
}

sub importStatusFile
{
	my $self = shift;
	my $statusfile = shift;
	
	# import HOST status lines
	foreach my $line (@{$statusfile->{HOSTS}})
	{
		warn "host_name not defined?\n" unless (defined $line->{FIELDS}->{host_name});
		my $hostServant = $self->getHost($line->{FIELDS}->{host_name});
		if (defined $hostServant)
		{
			$hostServant->{STATUS} = $line;
		}
	}
	
	# import SERVICE status lines
	foreach my $line (@{$statusfile->{SERVICES}})
	{
		my $servant = Config4GNU::Nagios::Service->new;
		$servant->{STATUS} = $line;

		$servant->{HOSTNAME} = $servant->getStat('host_name');
		$servant->{SERVICEDESCRIPTION}
			= $servant->getStat('service_description');
		$servant->{HOST} = $self->getHost($servant->{HOSTNAME});
		if (defined $servant->{HOST})
		{
			$servant->{HOST}->addService($servant);
		}
		else
		{
			warn "Host ", $servant->{HOSTNAME}, " not found";
		}
	}
}

sub getAllHostGroups
{
	my $self = shift;
	return $self->{hostgroups};
}

sub getHostGroup
{
	my $self = shift;
	my $targetName = shift;

	foreach my $hostgroup (@{$self->{hostgroups}})
	{
		return $hostgroup if ($hostgroup->name eq $targetName);
	}

	# not found
	return undef;
}

sub getHost
{
	my $self = shift;
	my $targetName = shift;
	
	foreach my $host (@{$self->{hosts}})
	{
		return $host if ($host->name eq $targetName);
	}

	# not found
	return undef;
}

package Config4GNU::Nagios::HostGroup;

sub new
{
	my $class = shift;
	my $objectDefinition = shift;
	
	my $self = {
		'CONFIGDEFINITION' => $objectDefinition };
	bless $self, $class;
	return $self;
}

sub name
{
	my $self = shift;
	
	return $self->{CONFIGDEFINITION}->getPropertyValue('hostgroup_name');
}

sub getProperty
{
	my $self = shift;
	my $propName = shift;

	return $self->{CONFIGDEFINITION}->getPropertyValue($propName);
}

sub getMemberHosts
{
	my $self = shift;
	my $main = $self->{MAIN} or die;
	
	my $members = $self->{CONFIGDEFINITION}->getPropertyValue('members');
	my @result;
	foreach my $hostname (split /,\s*/, $members)
	{
		push @result, $main->getHost($hostname);
	}
	return \@result;
}

package Config4GNU::Nagios::Host;

sub new
{
	my $class = shift;
	my ($objectDefinition, $statusLine) = @_;
	
	my $self = {
		'CONFIGDEFINITION' => $objectDefinition,
		'STATUS' => $statusLine };
	bless $self, $class;
	return $self;
}

sub name
{
	my $self = shift;
	
	return $self->{CONFIGDEFINITION}->getPropertyValue('host_name');
}

sub status
{
	my $self = shift;
	
	if (defined $self->{STATUS})
	{
		return $self->{STATUS}->{FIELDS}->{status};
	}
	return '';
}

sub address
{
	my $self = shift;
	return $self->{CONFIGDEFINITION}->getPropertyValue('address');
}

sub getProperty
{
	my $self = shift;
	my $propName = shift;

	return $self->{CONFIGDEFINITION}->getPropertyValue($propName);
}

sub getStat
{
	my $self = shift;
	my $statName = shift;

	if (defined $self->{STATUS})
	{
		return $self->{STATUS}->{FIELDS}->{$statName};
	}
	return "";
}

sub addService
{
	my $self = shift;
	my ($service) = @_;

	push @{$self->{SERVICES}}, $service;
}

sub getService
{
	my $self = shift;
	my $servicedescription = shift;

	foreach my $serv ($self->getServices)
	{
		if ($serv->{SERVICEDESCRIPTION} eq $servicedescription)
		{
			return $serv;
		}
	}
	return undef;
}

sub getServices
{
	my $self = shift;
	return defined($self->{SERVICES}) ? @{$self->{SERVICES}} : ();
}

package Config4GNU::Nagios::Service;

sub new
{
	my $class = shift;
	my ($objectDefinition, $statusLine) = @_;
	
	my $self = {
		'CONFIGDEFINITION' => $objectDefinition,
		'STATUS' => $statusLine };
	bless $self, $class;
	return $self;
}

sub getStat
{
	my $self = shift;
	my $statName = shift;

	if (defined $self->{STATUS})
	{
		return $self->{STATUS}->{FIELDS}->{$statName};
	}
	return "";
}
