#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use Config4GNU::SimpleKeyedValues;
use strict;

package Config4GNU::Nagios::Config;

package Config4GNU::Nagios::Config::MainConfig;
use base 'Config4GNU::SimpleKeyedValues';

package Config4GNU::Nagios::Config::ObjectConfig;
use base 'Config4GNU::AbstractProvider';

sub init
{
    my $self = shift;

#    $self->{SECTIONS} = [];
#	$self->{sectionServants} = [];
	
    open INFILE, '<', $self->{FILENAME}
            or throw Config4GNU::FileIOException errorMessage => "$!";
    my $section;
    my $comments = '';
    while (my $line = <INFILE>)
    {
		# TODO: handle line extenders (backslash at the end of the line)
		
        # check for a comment
        if ($line =~ /^\s*#/)
        {
            # ignore the comment
            $comments .= $line;
        }
        # check for section header
        elsif ($line =~ /^(\s*define\s+)(\w+)(\s*\{\s*)$/)
        {
            my $definepart = $1;
			my $section_type = $2;
			chomp (my $bracepart = $3);
			
            # create the backend object
			$section = $self->createSection($section_type);
    		$section->{COMMENTS} = $comments;
			$section->{definepart} = $definepart;
			$section->{bracepart} = $bracepart;

			# reset comments
			$comments = '';

			# add the section
			$self->addChildObject($section);
        }
        # check for property assignment
        elsif ($line =~ /^(\s*)(\w+)(\s+)(.*)$/ && defined($section))
        {
            my $whitespace1 = $1;
			my $prop_name = $2;
			my $whitespace2 = $3;
            chomp (my $prop_value = $4);
			my $endcomment = '';
			if ($prop_value =~ /(\s*;.*)$/)
			{
				$endcomment = $1;
				$prop_value = substr($prop_value, 0, - length($endcomment));
			}

			# use the current section to create the property object
            my $prop = $section->createProperty($prop_name, $prop_value);
			$prop->{COMMENTS} = $comments;
			$prop->{ENDCOMMENT} = $endcomment;
			$prop->{whitespace1} = $whitespace1;
			$prop->{whitespace2} = $whitespace2;

			# reset comments
            $comments = '';

			# add the property to the current section
			$section->addPropertyObject($prop);
        }
		# check for end of section
		elsif ($line =~ /^(\s*\}\s*)$/ && defined($section))
		{
			$section->{ENDCOMMENT} = $comments . $line;
		}
        else
        {
            $comments .= $line;
        }
    }
    close INFILE;

    $self->{ENDCOMMENT} = $comments;
}

sub createSection
{
	my $self = shift;
	my ($section_type) = @_;

    return Config4GNU::Nagios::Config::ObjectSection->new($section_type);
}

sub save
{
    my $self = shift;

	open OUTFILE, '>', $self->{FILENAME}
			or throw Config4GNU::FileIOException errorMessage => "$!";

    my @sections = @{$self->children};
    for my $section (@sections)
    {
        $section->writeConfig(\*OUTFILE);
    }
    print OUTFILE $self->{ENDCOMMENT};
	
	close OUTFILE;
}

package Config4GNU::Nagios::Config::ObjectSection;
use base 'Config4GNU::AbstractNode';

# create a new instance with the specified name
sub new
{
    my $class = shift;
    my $self = {};
		
	# arguments
	my ($object_type) = @_;

    $self->{TYPE} = $object_type;
	bless $self, $class;
	
	$self->init;
    return $self;
}

sub init
{
	my $self = shift;
#    $self->{PROPERTIES} = [];
#    $self->{propertyServants} = [];
}

sub name
{
    my $self = shift;
    return $self->{TYPE};
}

sub createProperty
{
	my $self = shift;
	my ($prop_name, $prop_value) = @_;
	
	return Config4GNU::Nagios::Config::ObjectProperty->new(
			$prop_name, $prop_value);
}

sub writeConfig
{
    my $self = shift;
	my $fh = shift;
	
    # sanity check parts of section
	# TODO: check comments, whitespace
	
	# output section comments and header
	print $fh $self->{COMMENTS};
    print $fh $self->{definepart}, $self->{TYPE}, $self->{bracepart}, "\n";

	# iterate through each property in this section
    foreach my $prop (@{$self->properties})
    {
		$prop->writeConfig($fh);
    }
	
	print $fh $self->{ENDCOMMENT};
}

# return TRUE if this object definition is a template
sub isTemplate
{
	my $self = shift;
	my $register = $self->getPropertyValue('register');
	return (defined($register) && $register eq '0');
}

sub getPropertyValue
{
	my $self = shift;
	my $targetName = shift;
	foreach my $prop (@{$self->properties})
	{
		if ($prop->{NAME} eq $targetName)
		{
			return $prop->{VALUE};
		}
	}
	return undef;
}

package Config4GNU::Nagios::Config::ObjectProperty;
use base 'Config4GNU::AbstractProperty';

sub writeConfig
{
    my $self = shift;
	my $fh = shift;
		
	# sanity check parts of each property
	# TODO: check comments
	#		for each line, should begin with valid comment character
	#		should end with a linefeed
	if (!defined($self->{COMMENTS}))
	{
		$self->{COMMENTS} = '';
	}
	if (!defined($self->{whitespace1}) || $self->{whitespace1} =~ /\S/)
	{
		# whitespace contains a nonwhitespace character, so reset it
		$self->{whitespace1} = '';
	}
	if (!defined($self->{ENDCOMMENT}) || $self->{ENDCOMMENT} !~ /^\s*;/)
	{
		# invalid comment string, so reset it
		$self->{ENDCOMMENT} = '';
	}

	# output the property
    print $fh $self->{COMMENTS};
    print $fh $self->{whitespace1}, $self->{NAME}, $self->{whitespace2},
			$self->{VALUE}, $self->{ENDCOMMENT}, "\n";
}

1;
