#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use Config4GNU::Ini;
use strict;

package Config4GNU::Samba;
use base 'Config4GNU::Ini';
$Config4GNU::Samba::DEFAULT_FILE = '/etc/samba/smb.conf';

=head1 NAME

Config4GNU::Samba - Perl interface for accessing/modifying Samba configuration
information.

=head1 SYNOPSIS

  use Config4GNU::Samba;
  my $samba = Config4GNU::Samba->new('/etc/samba/smb.conf');

  # change the WINS server setting
  my $global_section = $samba->getSection('global');
  my $oldvalue = $global_section->getPropertyValue('wins server');
  $global_section->setPropertyValue('wins server', '192.168.0.10');

  # add a new share
  my $tmp_share = $samba->newSection('tmp');
  $tmp_share->setPropertyValue('path', '/tmp');
  $tmp_share->setPropertyValue('read only', 'false');

  # delete a share
  $samba->deleteSection('tmp');

  # write the changes back to the file
  $samba->save;

=head1 DESCRIPTION

=head2 Methods of Config4GNU::Samba

=over

=item $samba = Config4GNU::Samba->new($file);

Parses the specified file as a smb.conf file and returns a data structure that
allows you to access and modify the settings contained in the file. Use the
C<save> method to write back the file. You may optionally not specify a file
in which case the default, '/etc/samba/smb.conf', will be used.

=cut

# create a new instance with the specified filename
sub new
{
    my ($class, $file) = @_;
	unless (defined $file)
	{
		$file = $Config4GNU::Samba::DEFAULT_FILE;
	}
	return $class->SUPER::new($file);
}

=item $section = $samba->getSection($name)

Returns an object representing the specified section of the Samba configuration
file. This object is of type Config4GNU::Samba::Section.
If the section does not exist, C<undef> is returned.

=cut

sub getSection
{
	my $self = shift;
	my $name = shift;
	
	# iterate through each section in the config file
	foreach my $section (@{$self->{CHILDREN}})
	{
		if ($section->name eq $name)
		{
			# found specified section, so return it
			return $section;
		}
	}
	
	# not found
	return undef;
}

=item my $section = $samba->newSection($name);

Creates a new section with the name $name, and appends it to the
configuration file. The new section will be empty, i.e. no properties
are defined for you.

=cut

sub newSection
{
	my $self = shift;
	my ($section_name) = @_;

	my $newSection = $self->createSection($section_name);
	$self->addChildObject($newSection);
	return $newSection;
}

=item $samba->deleteSection($section_name);

Deletes the specified section from the configuration file.

=cut

sub deleteSection
{
	my $self = shift;
	my $section_name = shift;

	my $sectionToDelete = $self->getSection($section_name);
	if (defined $sectionToDelete)
	{
		$self->removeChild($sectionToDelete);
	}
	else
	{
		warn "$section_name not found";
	}
}

=item $samba->save

Writes the configuration file.

=cut

sub createSection
{
	my $self = shift;
	my ($section_name) = @_;

	if ($section_name eq 'global')
	{
		return Config4GNU::Samba::GlobalSection->new($section_name);
	}

	# create a normal section
    return Config4GNU::Samba::Section->new($section_name);
}

=item $samba->restartService

This is a convenience method for restarting the Samba daemons. It will
probably be refactored as a different API or removed from this module
altogether. Note that the location of the config file specified when the
Samba object is created has no bearing on how to restart the service.

=cut

sub restartService
{
	my $self = shift;
	
	my $output = `/etc/init.d/samba restart 2>&1`;
	my $exit_value = $? >> 8;
	
	return 1 if ($exit_value == 0);

	print STDERR "output: $output\n";
	print STDERR "exit_value: $exit_value\n";
	return '';
}

=back

=cut

package Config4GNU::Samba::Section;
use base 'Config4GNU::Ini::Section';

%Config4GNU::Samba::Section::BOOLEAN_PROPERTIES =
	(	'allow trusted domains' => 1,
		'available' => 1,
		'bind interfaces only' => 1,
		'blocking locks' => 1,
		'browseable' => 1,
		'browse list' => 1,
		'case sensitive' => 1,
		'debug hires timestamp' => 1,
		'debug pid' => 1,
		'debug uid' => 1,
		'default devmode' => 1,
		'delete readonly' => 1,
		'delete veto files' => 1,
		'disable spoolss' => 1,
		'dns proxy' => 1,
		'domain logons' => 1,
		'domain master' => 1,
		'dos filemode' => 1,
		'dos filetime resolution' => 1,
		'dos filetimes' => 1,
		'encrypt passwords' => 1,
		'enhanced browsing' => 1,
		'fake directory create times' => 1,
		'fake oplocks' => 1,
		'follow symlinks' => 1,
		'force unknown acl user' => 1,
		'getwd cache' => 1,
		'guest ok' => 1,
		'guest only' => 1,
		'hide dot files' => 1,
		'hide local users' => 1,
		'hide unreadable' => 1,
		'homedir map' => 1,
		'host msdfs' => 1,
		'inherit acls' => 1,
		'inherit permissions' => 1,
		'kernel oplocks' => 1,
		'lanman auth' => 1,
		'large readwrite' => 1,
		'level2 oplocks' => 1,
		'load printers' => 1,
		'local master' => 1,
		'locking' => 1,
		'mangle case' => 1,
		'mangled names' => 1,
		'map archive' => 1,
		'map hidden' => 1,
		'map system' => 1,
		'msdfs root' => 1,
		'nis homedir' => 1,
		'nt acl support' => 1,
		'nt pipe support' => 1,
		'nt smb support' => 1,
		'nt status support' => 1,
		'null passwords' => 1,
		'obey pam restrictions' => 1,
		'only user' => 1,
		'oplocks' => 1,
		'pam password change' => 1,
		'passwd chat debug' => 1,
		'posix locking' => 1,
		'postscript' => 1,
		'preexec close' => 1,
		'preferred master' => 1,
		'preserve case' => 1,
		'printable' => 1,
		'profile acls' => 1,
		'read bmpx' => 1,
		'read only' => 1,
		'read raw' => 1,
		'restrict anonymous' => 1,
		'root preexec close' => 1,
		'set directory' => 1,
		'share modes' => 1,
		'short preserve case' => 1,
		'show add printer wizard' => 1,
		'ssl' => 1,
		'ssl compatibility' => 1,
		'ssl require clientcert' => 1,
		'ssl require servercert' => 1,
		'stat cache' => 1,
		'strict allocate' => 1,
		'strict locking' => 1,
		'strict sync' => 1,
		'strip dot' => 1,
		'sync always' => 1,
		'syslog only' => 1,
		'time server' => 1,
		'unix extensions' => 1,
		'unix password sync' => 1,
		'update encrypted' => 1,
		'use client driver' => 1,
		'use mmap' => 1,
		'use rhosts' => 1,
		'use sendfile' => 1,
		'utmp' => 1,
		'wide links' => 1,
		'winbind enum users' => 1,
		'winbind enum groups' => 1,
		'winbind use default domain' => 1,
		'wins proxy' => 1,
		'wins support' => 1,
		'write raw' => 1
	);
%Config4GNU::Samba::Section::INVERTED_BOOLEAN_PROPERTIES =
	(	'writable' => 'read only',
		'writeable' => 'read only',
		'write ok' => 'read only' );
%Config4GNU::Samba::Section::SYNONYMS =
	(	'allow hosts' => 'hosts allow',
		'auto services' => 'preload',
		'browsable' => 'browseable',
		'casesignames' => 'case sensitive',
		'create mode' => 'create mask',
		'debuglevel' => 'log level',
		'default' => 'default service',
		'deny hosts' => 'hosts deny',
		'directory' => 'path',
		'directory mode' => 'directory mask',
		'exec' => 'preexec',
		'group' => 'force group',
		'lock dir' => 'lock directory',
		'min passwd length' => 'min password length',
		'only guest' => 'guest only',
		'prefered master' => 'preferred master',
		'print ok' => 'printable',
		'printcap' => 'printcap name',
		'printer' => 'printer name',
		'protocol' => 'max protocol',
		'public' => 'guest ok',
		'root' => 'root directory',
		'root dir' => 'root directory',
		'timestamp logs' => 'debug timestamp',
		'user' => 'username',
		'users' => 'username'
	);

sub createProperty
{
	my $self = shift;
	my ($prop_name, $prop_value) = @_;
	
	# check for boolean types
	my $canonical_name = $Config4GNU::Samba::Section::SYNONYMS{$prop_name};
	unless (defined $canonical_name)
	{
		$canonical_name = $prop_name;
	}

	if ($Config4GNU::Samba::Section::BOOLEAN_PROPERTIES{$canonical_name})
	{
		my $prop = Config4GNU::Samba::BooleanProperty->new($prop_name, $prop_value);
		if ($canonical_name ne $prop_name)
		{
			$prop->{'override_name'} = $canonical_name;
		}
		return $prop;
	}
	elsif (defined $Config4GNU::Samba::Section::INVERTED_BOOLEAN_PROPERTIES{$prop_name})
	{
		# samba's configuration file has this annoying "inverted" boolean property...
		# a synonym for 'read only'
		my $prop = Config4GNU::Samba::InvertedBooleanProperty->new(
				$prop_name, $prop_value);
	    $prop->{'override_name'} = $Config4GNU::Samba::Section::INVERTED_BOOLEAN_PROPERTIES{$prop_name};
		return $prop;
	}
	
	# default
	my $prop = Config4GNU::Samba::Property->new($prop_name, $prop_value);
	if ($canonical_name ne $prop_name)
	{
		$prop->{'override_name'} = $canonical_name;
	}
	return $prop;
}

=head2 Methods of Config4GNU::Samba::Section

=over

=item $value = $section->getPropertyValue($propertyname)

Retrieves the defined property's value and returns it, or C<undef> if the
property is not defined.

(Defined in Config4GNU::AbstractNode)

=item $section->setPropertyValue($propertyname, $newvalue)

Sets the value of the specified property. If the property is not defined,
it will be created. To undefine a property, specify C<undef> for $newvalue.

=cut

sub setPropertyValue
{
	my $self = shift;
	my ($propertyname, $newvalue) = @_;

	my $prop = $self->getProperty($propertyname);
	if (defined $prop)
	{
		if (defined $newvalue)
		{
			# replace value of existing property
			$prop->setValue($newvalue);
		}
		else
		{
			# property exists, but caller wants it undefined
			$self->removeProperty($prop);
		}
	}
	else
	{
		# create a new property
		my $prop = $self->createProperty($propertyname, $newvalue);
		$self->addPropertyObject($prop);
	}
}

=back

=cut

package Config4GNU::Samba::GlobalSection;
use base 'Config4GNU::Samba::Section';

package Config4GNU::Samba::Property;
use base 'Config4GNU::Ini::Property';

sub name
{
    my $self = shift;
	
	# with Samba's many synonym property names do the following...
	# create the property object using the name as found in the configuration
	# file...
	# then set $prop->{'override_name'} to the canonical name for the property

    return defined($self->{'override_name'}) ? $self->{'override_name'} : $self->{NAME};
}

package Config4GNU::Samba::BooleanProperty;
use base 'Config4GNU::Samba::Property';

@Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_TRUE = ('true', 'yes', '1');
@Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_FALSE = ('false', 'no', '0');

sub value
{
    my $self = shift;
	
	if (grep {$_ eq lc($self->{VALUE})} @Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_TRUE)
	{
		return $Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_TRUE[0];
	}
	elsif (grep {$_ eq lc($self->{VALUE})} @Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_FALSE)
	{
		return $Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_FALSE[0];
	}
    return $self->{VALUE};
}

sub setValue
{
	my $self = shift;
	my $newvalue = shift;

	if ((grep {$_ eq lc($newvalue)} @Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_TRUE) ||
	    (grep {$_ eq lc($newvalue)} @Config4GNU::Samba::BooleanProperty::ALLOWED_VALUES_FALSE))
	{
		$self->{VALUE} = $newvalue;
	}
	else
	{
		throw Config4GNU::Property::InvalidValue
			msg => 'Valid values are yes/no, true/false, 1/0';
	}
}

package Config4GNU::Samba::InvertedBooleanProperty;
use base 'Config4GNU::Samba::BooleanProperty';

sub value
{
    my $self = shift;
	my $value = $self->SUPER::value;
	
	# return the opposite of whatever is actually defined
	return $value eq 'true' ? 'false' :
		   $value eq 'false' ? 'true' : $value;
}

sub setValue
{
	my $self = shift;
	my $newvalue = shift;

	$self->SUPER::setValue($newvalue eq 'true' ? 'false' :
			$newvalue eq 'false' ? 'true' : $newvalue);
}

=head1 BUGS

This parser does not support smb.conf files with line continuations
or include files.

=head1 SEE ALSO

See the Config4GNU website at http://config4gnu.sourceforge.net/

=head1 AUTHOR

Jason Long <jason@long.name>

=cut

1;
