#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use Config4GNU::AbstractProvider;
use Config4GNU::AbstractProperty;
use Config4GNU::FileIOException;
use strict;

=head1 NAME

Config4GNU::Sectioned - a CFG provider for sectioned configuration files.

=head1 DESCRIPTION

The sectioned format is a line-based configuration file, with sections at
the top level, which can contain properties and nested sections.

=cut


package Config4GNU::Sectioned::Section;
use base 'Config4GNU::AbstractNode';

=head2 Config4GNU::Sectioned::Section

=head3 Properties inherited from AbstractNode

=over

=item $self->{NAME}

Name of the section/node.

=item $self->{COMMENTS}

=item $self->{whitespace}

=item $self->{CHILDREN}

=back

=head3 New Properties

=over

=item $self->{WORDS}

Reference to array of Word objects that make up the line that starts
this section.

e.g. if the section starts with "Section InputDevice"
then $self->{WORDS} contains [ "Section", "InputDevice" ]

=item $self->{endCOMMENTS}

=item $self->{endWhitespace}

=item $self->{endWords}

=back

=cut

# create a new instance with the specified name
sub new
{
    my $class = shift;
    my $self = {};
	my @words = @_;

    $self->{NAME} = $words[1]->word;
	$self->{WORDS} = \@words;
	bless $self, $class;

	$self->{endCOMMENTS} = '';
	$self->{endWhitespace} = '';
	$self->{endWords} = undef;
	
	$self->init;
    return $self;
}

sub writeConfig
{
    my $self = shift;
	my $fh = shift;
	
    # sanity check parts of section
	# TODO: check comments, whitespace
	
	# output section comments and header
	print $fh $self->{COMMENTS};
    print $fh $self->{whitespace};
	foreach my $word (@{$self->{WORDS}})
	{
		print $fh $word->quoted, $word->{whitespace};
	}
	print $fh "\n";
	
	# iterate through each property in this section
    foreach my $child (@{$self->{CHILDREN}})
    {
		$child->writeConfig($fh);
    }
	
	print $fh $self->{endCOMMENTS};
	print $fh $self->{endWhitespace};

	# print end tokens, if any
	if (defined $self->{endWords})
	{
		foreach my $word (@{$self->{endWords}})
		{
			print $fh $word->quoted, $word->{whitespace};
		}
		print $fh "\n";
	}
}

sub addSectionEnd
{
	my $self = shift;
	my @words = @_;
	
	$self->{endWords} = \@words;
}

package Config4GNU::Sectioned;
use base ('Config4GNU::AbstractProvider', 'Config4GNU::Sectioned::Section');

sub init
{
    my $self = shift;
	$self->SUPER::init;

    open INFILE, '<', $self->{FILENAME}
            or throw Config4GNU::FileIOException errorMessage => "$!";
    my @section_stack = ($self);
    my $comments = '';
	
    while (my $line = <INFILE>)
    {
		# TODO: handle line extenders (backslash at the end of the line)
		
        # check for a comment
        if ($line =~ /^\s*[#;]/)
        {
            # ignore the comment
            $comments .= $line;
        }
		elsif ($line =~ /^(\s*)(\S.*)$/)
		{
			# break line into "words"
			my $whitespace = $1;
			my @words = $self->words($2);
			
			# check for start of a section
			if ($section_stack[0]->isSectionStart(@words))
			{
				# create the section object
				my $section = $section_stack[0]->createSection(@words);
				$section->{COMMENTS} = $comments;
				$section->{whitespace} = $whitespace;
				
				# reset comments
				$comments = '';
				
				# add the section
				$section_stack[0]->addChildObject($section);
				unshift @section_stack, $section;
			}
			# check for end of a section
			elsif ($section_stack[0]->isSectionEnd(@words))
			{
				$section_stack[0]->addSectionEnd(@words);
				$section_stack[0]->{endCOMMENTS} = $comments;
				$section_stack[0]->{endWhitespace} = $whitespace;
				
				# reset comments
				$comments = '';
				
				# pop off the section stack
				shift @section_stack;
				unless (defined $section_stack[0])
				{
					$section_stack[0] = $self;
				}
			}
			# assume it's a property
			else
			{
				my $prop = $section_stack[0]->createProperty(@words);
				$prop->{COMMENTS} = $comments;
				$prop->{whitespace} = $whitespace;
				
				# reset comments
				$comments = '';
				
				# add the property to the current section
				$section_stack[0]->addPropertyObject($prop);
			}
        }
        else
        {
            $comments .= $line;
        }
    }
    close INFILE;

    $self->{END_COMMENTS} = $comments;
}

sub save
{
    my $self = shift;

	open OUTFILE, '>', $self->{FILENAME}
			or throw Config4GNU::FileIOException errorMessage => "$!";

	# iterate through each property/section in this section
	foreach my $child (@{$self->{CHILDREN}})
	{
		$child->writeConfig(\*OUTFILE);
	}
	print OUTFILE $self->{END_COMMENTS};

	close OUTFILE;
}

sub words
{
	my $self = shift;
	my $line = shift;

	my $tokenizerClass = $self->TokenizerClass;
	my $tokenizer = $tokenizerClass->new($line);
	my @words = ();
	while (defined(my $word = $tokenizer->nextToken))
	{
		push @words, $word;
	}
	return @words;
}

sub TokenizerClass
{
	return 'Config4GNU::Sectioned::LineTokenizer';
}

package Config4GNU::Sectioned::Property;
use base ('Config4GNU::AbstractProperty');

sub new
{
	my $class = shift;
	my @words = @_;

	my $prop_name = $words[0]->word;
	my $self = $class->SUPER::new($prop_name);
	$self->{WORDS} = \@words;
	return $self;
}

sub writeConfig
{
    my $self = shift;
	my $fh = shift;

	# sanity check parts of each property
	# TODO: check comments
	#		for each line, should begin with valid comment character
	#		should end with a linefeed
	if (!defined($self->{comments}))
	{
		$self->{comments} = '';
	}
	if (!defined($self->{whitespace}) || $self->{whitespace} =~ /\S/)
	{
		# whitespace contains a nonwhitespace character, so reset it
		$self->{whitespace} = '';
	}
	if (!defined($self->{equalspart}) || $self->{equalspart} !~ /^\s*=\s*$/)
	{
		# invalid comment string, so reset it
		$self->{equalspart} = ' = ';
	}

	# output the property
    print $fh $self->{COMMENTS};
    print $fh $self->{whitespace};
	foreach my $word (@{$self->{WORDS}})
	{
		print $fh $word->quoted, $word->{whitespace};
	}
	print $fh "\n";
}

# return stringified value
sub value
{
    my $self = shift;
	my @result = ();
	foreach my $word (@{$self->{WORDS}})
	{
		push @result, $word->word;
	}
	shift @result;
    return join(',', @result);
}

package Config4GNU::Sectioned::LineTokenizer;

sub new
{
	my $class = shift;
	my $line = shift;
	
	my $self = {
		'LINE' => $line
	};
	bless $self, $class;
	return $self;
}

# split string into words
# returns undef if no more words
sub nextToken
{
	my $self = shift;
	my $line = $self->{LINE};

	# return undef if no more words
	return undef if ($line eq '');
		
	my $curWord = '';
	my $inQuote = '';
	my $whitespace = '';
	
	while ($line =~ /^./)
	{
		my $c = $&;
		$line = $';
		
		if ($c eq '"' || $c eq "'")
		{
			if ($inQuote eq $c)
			{
				$inQuote = '';
			}
			elsif ($inQuote eq '')
			{
				$inQuote = $c;
			}
		}
		elsif ($c eq '\\' && length($line) >= 1)
		{
			# backslash... include the following character literally
			$c .= substr($line, 0, 1);
			$line = substr($line, 1);
		}
		elsif ($c =~ /\s/ && $inQuote eq '')
		{
			# found whitespace... suck up the rest of the whitespace,
			# and add this word to the resulting array
			$line =~ /^\s*/;
			$whitespace = $c . $&;
			$line = $';
			
			last;
		}
		$curWord .= $c;
	}
	$self->{LINE} = $line;
	return Config4GNU::Sectioned::Word->new($curWord, $whitespace);
}

package Config4GNU::Sectioned::Word;

sub new
{
	my $class = shift;
	my $quoted = shift;
	my $whitespace = shift;
	
	my $self = {
		'quoted' => $quoted,
		'whitespace' => $whitespace
	};
	bless $self, $class;
	return $self;
}

sub quoted
{
	my $self = shift;
	return $self->{quoted};
}

sub newQuotes
{
	my $self = shift;
	my $word = $self->word;
	$word =~ s!\\!\\\\!g;
	$word =~ s!"!\\"!g;
	return "\"$word\"";
}

sub word
{
	my $self = shift;
	my $word = $self->{quoted};
	# remove surrounding quotes
	$word =~ s/^"(.*)"$/$1/e;
	# remove backslashes
	$word =~ s!\\"!"!g;
	$word =~ s!\\\\!\\!g;
	return $word;
}

1;
