#!/usr/bin/perl

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
use warnings;

=head1 NAME

Config4GNU::Tokenizer - read tokens from an open configuration file

=head1 DESCRIPTION

=cut

package Config4GNU::Tokenizer;

=head2 Config4GNU::Tokenizer

=head3 Methods

=over

=item my $tokenizer = Config4GNU::Tokenizer->new($filehandle);

Given a file handle that is an open file, this returns an object that
lets you read from that open file one "token" at a time.

=cut

sub new
{
	my $class = shift;
	my $filehandle = shift;
	
	my $self = {
		'FH' => $filehandle };
	return bless $self, $class;
}

sub openFile
{
	my $class = shift;
	my $filename = shift;

	my $fh;
	open $fh, '<', $filename
		or die "Cannot open $filename: $!\n";
	return $class->new($fh);
}

=item my $token = $tokenizer->nextToken;

Reads a "token" from the file and returns it. If at the end of the file, the
C<undef> value is returned. The meaning of "token" is dependent on the
implementation.

=cut

sub nextToken
{
	my $self = shift;
	
	# if the file is already closed, return undef
	unless (defined $self->{'FH'})
	{
		return undef;
	}
	
	my $curLine = $self->{'curLine'};
	if (!defined($curLine) || $curLine eq '')
	{
		$curLine = $self->readLine;
		return undef unless (defined $curLine);
	}
	
	my $token;
	if ($curLine =~ /^\s+/)
	{
		# whitespace
		$token = $&;
		$curLine = $';
	}
	elsif ($curLine =~ /^;/)
	{
		# comment ... continues until end of line
		$token = $curLine;
		$curLine = undef;
	}
	elsif ($curLine =~ /^\$\w+/)
	{
		# control entry (e.g. $ORIGIN, $INCLUDE, $TTL)
		$token = $&;
		$curLine = $';
	}
	elsif ($curLine =~ /^"/)
	{
		# begin quoted character string
		$token = $&;
		$curLine = $';
		while (1)
		{
			if (!defined($curLine) || $curLine eq '')
			{
				$curLine = $self->readLine;
				if (!defined($curLine))
				{
					return $token;
				}
			}
			if ($curLine =~ /^\\./)
			{
				$token .= $&;
				$curLine = $';
			}
			elsif ($curLine =~ /^"/)
			{
				$token .= $&;
				$curLine = $';
				last;
			}
			elsif ($curLine =~ /^[^\\"]+/)
			{
				$token .= $&;
				$curLine = $';
			}
			else
			{
				warn "Unknown string literal $curLine";
				last;
			}
		}
	}
	elsif ($curLine =~ /^[-\w.]+/)
	{
		# unquoted character string
		$token = $&;
		$curLine = $';
	}
	elsif ($curLine =~ /^[@)(]/)
	{
		# single-character token
		$token = $&;
		$curLine = $';
	}
	else
	{
		warn "Unknown token $curLine";
		return undef;
	}
	
	$self->{'curLine'} = $curLine;
	return $token;
}

sub readLine
{
	my $self = shift;

	my $fh = $self->{'FH'};
	my $line = <$fh>;
	unless (defined $line)
	{
		close $self->{'FH'};
		return $self->{'FH'} = undef;
	}
	return $line;
}

=back

=head3 Properties

=over

=item FH

The file handle for the open file. Becomes undefined when the end of the
input file is reached.

=item curLine

A string containing the current line that is being read.

=back

=head1 AUTHOR

Jason Long <jason@long.name>

=cut

1;
