#!/usr/bin/perl -w

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use owperl;
use strict;
use warnings;
use Error qw(:try);

package main;

%main::providers = ();

sub registerProvider
{
	my ($cim_class, $perl_class) = @_;
	$main::providers{lc($cim_class)} = $perl_class;
	print STDERR "Registered provider for $cim_class\n";
}

sub enumInstances
{
	my ($env, $ns, $className, $localOnly, $deep, $includeQualifiers,
	  $includeClassOrigin, $result, $cimClass, $requestedClass) = @_;
	
	print "Entering main::enumInstances\n";

	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		$providerclass->enumInstances(@_);
	}
	else
	{
		die("Unrecognized class $className");
	}
}

sub enumInstanceNames
{
    my ($env, $ns, $className, $result, $cimClass) = @_;
	
	print "Entering main::enumInstanceNames\n";

	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		$providerclass->enumInstanceNames(@_);
	}
	else
	{
		die("Unrecognized class $className");
	}
}

sub getInstance
{
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	print "Entering main::getInstance\n";

	# what class is requested?
	my $className = $cimClass->getName;
	print "  className: $className\n";
	print "  instanceName: ", $instanceName->toString, "\n";

	# I noticed here that instanceName always has a blank namespace
	# here (and hostname as well but I haven't cared about that yet).
	# Most of the time this doesn't matter, but if you're constructing
	# other object paths (i.e. the associatorNames method) make sure
	# you use the provided $ns rather than relying on the $instanceName
	# variable to have the correct namespace.

	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		return $providerclass->getInstance(@_);
	}
	else
	{
		die("Unrecognized class $className");
		return undef;
	}
}

sub associators
{
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole, $includeQualifiers, $includeClassOrigin,
		$propertyList) = @_;
	
	print "Entering main::associators\n";

	my $providerclass = $main::providers{lc($assocClass)};
	if (defined $providerclass)
	{
		$providerclass->associators(@_);
	}
	else
	{
		die("Unrecognized class $assocClass");
	}
}

sub associatorNames
{
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole) = @_;
	
	print "Entering main::associatorNames\n";
	
	my $providerclass = $main::providers{lc($assocClass)};
	if (defined $providerclass)
	{
		$providerclass->associatorNames(@_);
	}
	else
	{
		die("Unrecognized class $assocClass");
	}
}

sub references
{
	my ($env, $result, $ns, $objectName, $resultClass, $role,
		$includeQualifiers, $includeClassOrigin, $propertyList) = @_;
	
	print "Entering main::references\n";
	print "  objectName = ", $objectName->toString, "\n";
	print "  resultClass = ", $resultClass, "\n";
	print "  role = ", $role, "\n";
	print "\n";

	my $providerclass = $main::providers{lc($resultClass)};
	if (defined $providerclass)
	{
		$providerclass->references(@_);
	}
	else
	{
		die("Unrecognized class $resultClass");
	}
}

sub referenceNames
{
	my ($env, $result, $ns, $objectName, $resultClass, $role) = @_;
	
	print "Entering main::referenceNames\n";
	print "  objectName = ", $objectName->toString, "\n";
	print "  resultClass = ", $resultClass, "\n";
	print "  role = ", $role, "\n";
	print "\n";
	
	my $providerclass = $main::providers{lc($resultClass)};
	if (defined $providerclass)
	{
		$providerclass->referenceNames(@_);
	}
	else
	{
		die("Unrecognized class $resultClass");
	}
}

sub createInstance
{
    my ($env, $ns, $cimInstance) = @_;
	
	print "Entering main::createInstance\n";

	my $className = $cimInstance->getClassName;
	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		return $providerclass->createInstance(@_);
	}
	else
	{
		die("Unrecognized class $className");
		return undef;
	}
}

sub modifyInstance
{
	my ($env, $ns, $modifiedInstance, $previousInstance,
		$includeQualifiers, $propertyList, $theClass) = @_;

	print "Entering main::modifyInstance\n";

	# what class is requested?
	my $className = $theClass->getName;

	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		$providerclass->modifyInstance(@_);
	}
	else
	{
		die ("Unrecognized class $className");
	}
}

sub deleteInstance
{
	my ($env, $ns, $cop) = @_;
	
	print "Entering main::deleteInstance\n";

	my $className = $cop->getClassName;
	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		$providerclass->deleteInstance(@_);
	}
	else
	{
		die("Unrecognized class $className");
	}
}

sub invokeMethod
{
	my ($env, $ns, $objectName, $methodName, $inParams, $outParams) = @_;
			  
	print "Entering main::invokeMethod\n";

	my $className = $objectName->getClassName;
	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		return $providerclass->invokeMethod(@_);
	}
	else
	{
		die "Unrecognized class $className";
		return undef;
	}
}

my %CIM_CLASS_CACHE = ();
sub get_cim_class
{
	my ($cimom, $ns, $className) = @_;
	if (not defined $CIM_CLASS_CACHE{$className})
	{
		$CIM_CLASS_CACHE{$className} = $cimom->getClass($ns, $className);
	}
	return $CIM_CLASS_CACHE{$className};
}
########################################################################

package Config4GNU::WBEM::ProviderException;
use base 'Error';

sub new
{
	my $self = shift;
	my %args = @_;
	
	local $Error::Depth = $Error::Depth + 1;
	$self->SUPER::new(@_);
}

########################################################################
# Provider is the base class for all Config4GNU::WBEM style providers.
# It provides a "register" method that associates itself with a
# particular CIM class. It also provides stub methods that throw
# exceptions when they're called. It is recommended to use the
# SimpleProvider class rather than this class.

package Config4GNU::WBEM::Provider;

sub new
{
	my $class = shift;
	my %args = @_;
	
	my $provider = \%args;
	return bless $provider, $class;
}

sub register
{
	my $perl_class = shift;
	my $cim_class = shift;
	
	main::registerProvider($cim_class, $perl_class);
}

sub enumInstances
{
	my $classname = shift;
	die "EnumInstances not supported for $classname\n";
}

# default implementation of enumInstanceNames calls enumInstances and creates
# a list of names of the returned instances
sub enumInstanceNames
{
	my $self = shift;
    my ($env, $ns, $className, $result, $cimClass) = @_;

	my $resulttranslator = WBEM::Result2Path->new($ns, $result);
	$self->enumInstances($env, $ns, $className, 0, 0, 0, 0,
		$resulttranslator, $cimClass, $className);
}

sub getInstance
{
	my $providerclass = shift;
	my ($class, $cop) = @_;

	throw Config4GNU::WBEM::ProviderException -text =>
			"getInstance not supported for $providerclass";
}

sub associators
{
	my $providerclass = shift;

	die "Associators not supported for $providerclass\n";
}

sub associatorNames
{
	my $providerclass = shift;

	die "AssociatorNames not supported for $providerclass\n";
}

sub references
{
	my $providerclass = shift;

	die "References not supported for $providerclass\n";
}

sub referenceNames
{
	my $providerclass = shift;

	die "ReferenceNames not supported for $providerclass\n";
}

sub createInstance
{
	my $providerclass = shift;

	die "CreateInstance not supported for $providerclass\n";
}

sub modifyInstance
{
	my $providerclass = shift;

	die "ModifyInstance not supported for $providerclass\n";
}

sub deleteInstance
{
	my $providerclass = shift;

	die "DeleteInstance not supported for $providerclass\n";
}

sub is_subclass
{
	my $providerclass = shift;
	my ($env, $ns, $subclass, $superclass) = @_;
	print "is_subclass $subclass $superclass\n";
	unless ($subclass)
	{
		return 0;
	}
	if (lc($subclass) eq lc($superclass))
	{
		return 1;
	}
	# get parent class of subclass
	my $cimClass = main::get_cim_class($env->getCIMOMHandle, $ns, $subclass);
	return $providerclass->is_subclass($env, $ns,
			$cimClass->getSuperClass, $superclass);
}

########################################################################
# SimpleProvider implements enumInstances and associators by calling
# enumInstanceNames or associatorNames respectively to get a list of
# object paths, then calls the appropriate getInstance method for each
# name returned. The ResultMarshaller class is used to call the
# appropriate getInstance method.

package Config4GNU::WBEM::SimpleProvider;
use base 'Config4GNU::WBEM::Provider';

sub enumInstances
{
	my $providerclass = shift;
	my ($env, $ns, $className, $localOnly, $deep, $includeQualifiers,
		$includeClassOrigin, $result, $cimClass, $requestedClass) = @_;

	my $result1 = Config4GNU::WBEM::ResultMarshaller->new(
		ENV => $env,
		HANDLER => $result,
		NS => $ns,
		LOCAL_ONLY => $localOnly,
		DEEP => $deep,
		INCLUDE_QUALIFIERS => $includeQualifiers,
		INCLUDE_CLASS_ORIGIN => $includeClassOrigin);
	$providerclass->enumInstanceNames($env, $ns, $className, $result1,
		$cimClass);
}

sub enumInstanceNames
{
	my $classname = shift;
	die "EnumInstanceNames not supported for $classname\n";
}

sub associators
{
	my $providerclass = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole, $includeQualifiers, $includeClassOrigin,
		$propertyList) = @_;

	my $result1 = Config4GNU::WBEM::ResultMarshaller->new(
		ENV => $env,
		HANDLER => $result,
		NS => $ns,
		LOCAL_ONLY => 0,
		INCLUDE_QUALIFIERS => $includeQualifiers,
		INCLUDE_CLASS_ORIGIN => $includeClassOrigin,
		PROPERTY_LIST => $propertyList);
	$providerclass->associatorNames($env, $result1, $ns, $objectName,
		$assocClass, $resultClass, $role, $resultRole);
}

sub associatorNames
{
	my $classname = shift;
	die "AssociatorNames not supported for $classname\n";
}

sub references
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $resultClass, $role,
		$includeQualifiers, $includeClassOrigin, $propertyList) = @_;

	my $result1 = Config4GNU::WBEM::ResultMarshaller->new(
		ENV => $env,
		HANDLER => $result,
		NS => $ns,
		INCLUDE_QUALIFIERS => $includeQualifiers,
		INCLUDE_CLASS_ORIGIN => $includeClassOrigin,
		PROPERTY_LIST => $propertyList);
	$provider->referenceNames($env, $result1, $ns, $objectName,
		$resultClass, $role);
}

sub referenceNames
{
	my $classname = shift;
	die "ReferenceNames not supported for $classname\n";
}

#######################################################################
# ResultMarshaller acts as a proxying CIMObjectPathHandler that
# accesses the given instance and passes it to a CIMInstanceHandler.
# It optimizes for locally provided instances by accessing the
# local Perl provider registration. If it's a local provider, it calls
# the appropriate provider method directly, instead of going through
# the CIMOM. If not a local provider, it will call the CIMOM, which
# works for both repository-backed instance data and dynamic data.

package Config4GNU::WBEM::ResultMarshaller;

sub new
{
	my $class = shift;
	my %params = @_;
	return bless \%params, $class;
}

sub handle
{
	my $self = shift;
	my $instanceName = shift;

	# get the instance
	my $env = $self->{ENV};
	my $ns = $self->{NS};
	my $ci;

	# is this class type handled by this provider?
	my $className = $instanceName->getClassName;
	my $providerclass = $main::providers{lc($className)};
	if (defined $providerclass)
	{
		print STDERR "ResultMarshaller: $className is self-provided\n";
		my $cimClass = main::get_cim_class($env->getCIMOMHandle,
				$ns, $className);
		$ci = $providerclass->getInstance($env, $ns, $instanceName,
			$self->{LOCAL_ONLY}, $self->{INCLUDE_QUALIFIERS},
			$self->{INCLUDE_CLASS_ORIGIN}, $self->{PROPERTY_LIST},
			$cimClass);
	}
	else
	{
		# this works as long as the instance we're getting is not dynamic
		# if it is dynamic, I don't know what will happen
		$ci = $env->getCIMOMHandle->getInstance($ns, $instanceName);
	}
	if (defined $ci)
	{
		$self->{HANDLER}->handle($ci);
	}
	else
	{
		warn "Undefined CIMInstance returned from getInstance";
	}
}

#######################################################################
#
# To use this class, please implement getKeyInfo

package Config4GNU::WBEM::SimpleAssociationProvider;
use base 'Config4GNU::WBEM::SimpleProvider';

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;
	
	my $ci = $cimClass->newInstance;
	$provider->copyKeys($ci, $instanceName);
	return $ci;
}

sub copyKeys
{
	my $self = shift;
	my ($ci, $cop) = @_;
	my @k = $self->getKeyInfo;
	while (@k)
	{
		my $keyDataType = shift @k;
		my $keyName = shift @k;
		if ($keyDataType eq 'ref')
		{
			$ci->setRefProperty($keyName, $cop->getRefKeyValue($keyName));
		}
		else
		{
			die "Unknown key data type $keyDataType";
		}
	}
}

sub associatorNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole) = @_;
	
	my @k0 = $provider->getKeyInfo;
	my @k1 = ();
	while (@k0)
	{
		my $keyDataType = shift @k0;
		my $keyName = shift @k0;
		if ($keyDataType eq 'ref')
		{
			push @k1, $keyName;
		}
	}
	
	my $resultmarshaller = new
		Config4GNU::WBEM::SimpleAssociationProvider::ResultMarshaller
			( HANDLER => $result,
			  REFKEYS => \@k1,
			  ORIGIN => $objectName );
	$provider->referenceNames($env, $resultmarshaller, $ns,
		$objectName, $assocClass);
}

package Config4GNU::WBEM::SimpleAssociationProvider::ResultMarshaller;

sub new
{
	my $class = shift;
	my %params = @_;
	return bless \%params, $class;
}

sub handle
{
	my $self = shift;
	my $referenceName = shift;
	
	my @k = @{$self->{REFKEYS}};
	my $cop = $self->{ORIGIN};
	foreach (@k)
	{
		my $mcop = $referenceName->getRefKeyValue($_);
		next if ($mcop->equals($cop));
		$self->{HANDLER}->handle($mcop);
	}
}

#######################################################################
# SecondaryProvider should eventually allow you to implement a
# property provider. Instances are stored and accessed from the
# repository... the property provider only provides for certain
# properties, not creation/deletion of instances.

package Config4GNU::WBEM::SecondaryProvider;
use base 'Config4GNU::WBEM::SimpleProvider';

sub getInstance
{
	my $providerclass = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
		$includeClassOrigin, $propertyList, $cimClass) = @_;

	print "Entered SecondaryProvider::getInstance\n";
	# get instance from repository
	my $ci = $env->getRepositoryCIMOMHandle->getInstance(
			$ns, $instanceName, $localOnly, $includeQualifiers);
	#$providerclass->filterInstance($env, $ns, $ci);
	return $ci;
}

sub modifyInstance
{
	my $providerclass = shift;
	my ($env, $ns, $modifiedInstance, $previousInstance,
		$includeQualifiers, $propertyList, $theClass) = @_;
	
	print "Entered SecondaryProvider::modifyInstance\n";
	# modify the underlining data
	$env->getRepositoryCIMOMHandle->modifyInstance($ns, $modifiedInstance);
}

#######################################################################
# UltimateProvider is like a SecondaryProvider, except you can provide
# for instances not actually found in the repository. When an instance
# is requested that does not exist in the repository, a new instance is
# created and initialized with the keys of instanceName. Implement the
# copyKeys to initialize the key properties. Implement filterGetInstance
# to fill in the dynamic properties.
#
# When a modifyInstance call is made on an instance that is not in the
# repository, it is added instead. Implement filterModifyInstance to
# push out changes of dynamic data, and clear any data that shouldn't
# go to the repository.
#
# In summary, implement the following three methods. Do NOT override
# getInstance or modifyInstance.
#  copyKeys
#  filterGetInstance
#  filterModifyInstance

package Config4GNU::WBEM::UltimateProvider;
use base 'Config4GNU::WBEM::SecondaryProvider';

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
		$includeClassOrigin, $propertyList, $cimClass) = @_;
	print "Entered UltimateProvider::getInstance\n";

	my @args = @_;
	my $ci = eval { $provider->SUPER::getInstance(@args) };
	if ($@)
	{
		# exception occurred... presumably "instance not found"
		if ($@ =~ /requested object could not be found/)
		{
			# instance not found in repository, so create one
			$ci = $cimClass->newInstance;
			$provider->copyKeys($ci, $instanceName);
		}
		else # not "instance not found"
		{
			print STDERR "got exception from call to getInstance\n";
			print STDERR "error was $@\n";
			die $@;
		}
	}
	$provider->filterGetInstance($env, $ns, $instanceName, $ci);
	return $ci;
}

sub modifyInstance
{
	my $provider = shift;
	my ($env, $ns, $modifiedInstance, $previousInstance,
		$includeQualifiers, $propertyList, $theClass) = @_;
	
	print "Entered UltimateProvider::modifyInstance\n";

	$provider->filterModifyInstance($env, $ns,
			$modifiedInstance, $previousInstance);

	# update repository
	my @args = @_;
	eval { $provider->SUPER::modifyInstance(@args) };
	if ($@)
	{
		print "Could not modify instance, try creating instance instead\n";
		# exception occurred... presumably "instance not found"
		# so add the instance instead
		$env->getRepositoryCIMOMHandle->createInstance($ns, $modifiedInstance);
	}
}

#######################################################################

package WBEM::Result2Path;

sub new
{
	my ($class, $ns, $result) = @_;
	my $self = { 'NS' => $ns, 'RESULT' => $result };
	return bless $self, $class;
}

sub handle
{
	my $self = shift;
	my $ci = shift;

	my $cop = owperl::CIMObjectPath->new($self->{'NS'}, $ci);
	$self->{'RESULT'}->handle($cop);
}

1;
