#!/usr/bin/perl

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# USAGE:
#   Import the associated schema file (Nagios.mof) into the OpenWBEM
#   CIMOM using owmofc.
#   Install this file to $(OPENWBEM_PREFIX)/lib/openwbem/perlproviders/

use Config4GNU::WBEM;
use Config4GNU::Nagios;
use strict;
use warnings;

use constant CLASSNAME_NAGIOS => 'Config4GNU_NagiosService';
use constant CLASSNAME_HOSTED_NAGIOS => 'Config4GNU_NagiosHostedService';

# perform initialization tasks
NagiosService->register(CLASSNAME_NAGIOS);
HostedNagios->register(CLASSNAME_HOSTED_NAGIOS);
HostGroupDefinition->register('Config4GNU_NagiosHostGroupDefinition');
HostGroup->register('Config4GNU_NagiosHostGroup');
Host->register('Config4GNU_NagiosMonitoredHost');
HostInHostGroup->register('Config4GNU_NagiosHostInHostGroup');
HostService->register('Config4GNU_NagiosHostService');
Service->register('Config4GNU_NagiosMonitoredService');

my %nagios_objects = ();
my %nagios_lastupdates = ();

sub getNagios
{
	my $configfile = shift;

	my $curtime = time;
	if (!defined($nagios_objects{$configfile}) ||
		$nagios_lastupdates{$configfile} + 30 < $curtime)
	{
		$nagios_objects{$configfile} =
			Config4GNU::Nagios->new($configfile);
		$nagios_lastupdates{$configfile} = time;
	}
	return $nagios_objects{$configfile};
}

sub createNagiosObject
{
	my ($class) = @_;

	# create an instance for the first return value
	my $ci = $class->newInstance;

	# fill in the key properties
	$ci->setStringProperty(
			"ServiceName", "Nagios");
	$ci->setStringProperty(
			"ServiceCreationClassName", "Config4GNU_NagiosService");
	chomp (my $hostname = `hostname`);
	$ci->setStringProperty(
			"SystemName", $hostname);
	$ci->setStringProperty(
			"SystemCreationClassName", "CIM_System");

	return $ci;
}

sub time_pl2cim
{
	my $perltime = shift;

	my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) =
		gmtime($perltime);
	return sprintf("%04d%02d%02d%02d%02d%02d",
			$year + 1900, $mon + 1, $mday, $hour, $min, $sec)
			. ".000000+000";
}

#######################################################################
#######################################################################

package NagiosService;
use base 'Config4GNU::WBEM::SimpleProvider';
use constant CLASSNAME_NAGIOS => main::CLASSNAME_NAGIOS;

our @CONFIG_FILES;
@CONFIG_FILES = ('/usr/local/nagios/etc/nagios.cfg');

sub enumInstanceNames
{
	my $provider = shift;
	my ($env, $ns, $className, $result, $cimClass) = @_;

	my @config_files = grep(-f $_, @CONFIG_FILES);
	foreach my $config_file (@config_files)
	{
		my $mcop = owperl::CIMObjectPath->new(CLASSNAME_NAGIOS, $ns);
		$mcop->addStringKeyValue('Name', $config_file);
		$mcop->addStringKeyValue('CreationClassName', CLASSNAME_NAGIOS);
		chomp (my $hostname = `hostname`);
		$mcop->addStringKeyValue('SystemName', $hostname);
		$mcop->addStringKeyValue('SystemCreationClassName', "CIM_System");
		$result->handle($mcop);
	}
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $config_file = $instanceName->getStringKeyValue('Name');
	if (not -f $config_file)
	{
		die "Invalid Nagios service name $config_file\n";
	}

	my $ci = $cimClass->newInstance;

	# fill in the key properties
	$ci->setStringProperty('Name', $config_file);
	$ci->setStringProperty('CreationClassName', CLASSNAME_NAGIOS);
	$ci->setStringProperty('SystemName',
		$instanceName->getStringKeyValue('SystemName'));
	$ci->setStringProperty('SystemCreationClassName',
		$instanceName->getStringKeyValue('SystemCreationClassName'));

	# fill in other properties
	$ci->setStringProperty('ElementName', 'Nagios');
	$ci->setStringProperty('ConfigFile', $config_file);
	$ci->setBooleanProperty('DaemonMode', 1);

	return $ci;
}

#######################################################################

package HostedNagios;
use base 'Config4GNU::WBEM::SimpleAssociationProvider';

sub getKeyInfo
{
	return (
		'ref', 'Antecedent',
		'ref', 'Dependent'
	);
}

sub referenceNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $resultClass) = @_;

	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_NAGIOS))
	{
		my $mcop = owperl::CIMObjectPath->new(
			$objectName->getStringKeyValue('SystemCreationClassName'), $ns);
		$mcop->setStringKeyValue('Name',
			$objectName->getStringKeyValue('SystemName'));
		$mcop->setStringKeyValue('CreationClassName',
			$objectName->getStringKeyValue('SystemCreationClassName'));
		
		my $cop = owperl::CIMObjectPath->new(
				main::CLASSNAME_HOSTED_NAGIOS, $ns);
		$cop->setRefKeyValue('Antecedent', $mcop);
		$cop->setRefKeyValue('Dependent', $objectName);
		$result->handle($cop);
	}
	elsif ($provider->is_subclass($env, $ns, $baseClass, 'CIM_System'))
	{
		my $config_file = $CONFIG_FILES[0];
		my $mcop = owperl::CIMObjectPath->new(main::CLASSNAME_NAGIOS, $ns);
		$mcop->setStringKeyValue('Name', $config_file);
		$mcop->setStringKeyValue('CreationClassName', main::CLASSNAME_NAGIOS);
		$mcop->setStringKeyValue('SystemName',
			$objectName->getStringKeyValue('SystemName'));
		$mcop->setStringKeyValue('SystemCreationClassName',
			$objectName->getStringKeyValue('CreationClassName'));
		
		my $cop = owperl::CIMObjectPath->new(
				main::CLASSNAME_HOSTED_NAGIOS, $ns);
		$cop->setRefKeyValue('Antecedent', $objectName);
		$cop->setRefKeyValue('Dependent', $mcop);		
		$result->handle($cop);
	}
}

#######################################################################
##                                                                   ##
##                   NAGIOS OBJECT DEFINITIONS                       ##
##                                                                   ##
#######################################################################

package HostGroupDefinition;
use base 'Config4GNU::WBEM::Provider';

sub enumInstances
{
	my $self = shift;
	my ($env, $ns, $className, $localOnly, $deep, $includeQualifiers,
	  $includeClassOrigin, $result, $cimClass, $requestedClass) = @_;

	my $nagios = main::getNagios();

	foreach my $hostgroup (@{$nagios->getAllHostGroups})
	{
		# create an instance
		my $ci = main::createNagiosObject($cimClass);

		# fill in the remaining key properties
		$ci->setStringProperty('HostgroupName', $hostgroup->name);

		# fill in non-key properties
		$ci->setStringProperty('ElementName',
			$hostgroup->getProperty('alias'));

		$result->handle($ci);
	}
}

package HostDefinition;
use base 'Config4GNU::WBEM::Provider';

sub enumInstances
{
	my $self = shift;
	my ($env, $ns, $className, $localOnly, $deep, $includeQualifiers,
	  $includeClassOrigin, $result, $cimClass, $requestedClass) = @_;

	my $nagios = main::getNagios();
	my @result = ();

	foreach my $host (@{$nagios->{hosts}})
	{
		# create an instance
		my $ci = main::createNagiosObject($cimClass);

		# fill in the remaining key properties
		$ci->setStringProperty(
				"HostName", $host->name);

		# fill in non-key properties
		$ci->setStringProperty(
				"ElementName", $host->getProperty('alias'));
		$ci->setStringProperty(
				"Address", $host->getProperty('address'));

		$result->handle($ci);
	}
}

#######################################################################
##                                                                   ##
##                            HOST GROUP                             ##
##                                                                   ##
#######################################################################

package HostGroup;
use base 'Config4GNU::WBEM::Provider';

sub enumInstances
{
	my $self = shift;
	my ($env, $ns, $className, $localOnly, $deep, $includeQualifiers,
	  $includeClassOrigin, $result, $cimClass, $requestedClass) = @_;

	print STDERR "Entering HostGroup::enumInstances\n";

	my $nagios = main::getNagios();

	foreach my $hostgroup (@{$nagios->getAllHostGroups})
	{
		# create an instance
		my $ci = main::createNagiosObject($cimClass);

		# fill in the remaining key properties
		$ci->setStringProperty(
				"HostgroupName", $hostgroup->name);

		# fill in non-key properties
		$ci->setStringProperty(
				"ElementName", $hostgroup->getProperty('alias'));

		$result->handle($ci);
	}
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $nagios = main::getNagios();

	# find the specified host group
	my $targetName = $instanceName->getStringKeyValue('HostGroupName');
	unless (defined $targetName)
	{
		warn "CIMObjectPath is missing key value 'HostGroupName'";
		return $cimClass->newInstance;
	}

	my $hostgroup = $nagios->getHostGroup($targetName);
	if (!defined($hostgroup))
	{
		return undef;
	}

	# create a new instance and initialize with Nagios properties
	my $ci = main::createNagiosObject($cimClass);

	# fill in the remaining key properties
	$ci->setStringProperty(
			"HostGroupName", $hostgroup->name);

	# fill in non-key properties
	$ci->setStringProperty(
			"ElementName", $hostgroup->getProperty('alias'));

	return $ci;
}

#######################################################################
#######################################################################

package HostInHostGroup;
use base 'Config4GNU::WBEM::Provider';

sub associators
{
	my $providerclass = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole, $includeQualifiers, $includeClassOrigin,
		$propertyList) = @_;

	my $objectClass = $objectName->getClassName;
	print "ns = $ns\n";
	print "objectClass = $objectClass\n";

	# which class is requested?
	if (lc($objectClass) eq 'config4gnu_nagioshostgroup')
	{
		# what container is specified?
		my $containerName = $objectName->getStringKeyValue('HostGroupName');
		my $cimClass = $env->getCIMOMHandle()->getClass($ns,
				'Config4GNU_NagiosMonitoredHost');
		associators_listMembers($result, $containerName, $cimClass);
	}
	elsif (lc($objectClass) eq 'config4gnu_nagiosmonitoredhost')
	{
		# what member is specified?
		my $memberName = $objectName->getStringKeyValue('Name');
		my $cimClass = $env->getCIMOMHandle()->getClass($ns,
				'Config4GNU_NagiosHostGroup');
		associators_listContainers($result, $memberName, $cimClass);
	}
	else
	{
		die "Invalid classname $objectClass\n";
	}
}

sub associators_listMembers
{
	my ($result, $containerName, $cimClass) = @_;

	my $nagios = main::getNagios();

	# find the specified host group
	my $hostgroup = $nagios->getHostGroup($containerName);
	if (!defined($hostgroup))
	{
		# TODO: throw exception
		die "Hostgroup $containerName not found\n";
	}

	foreach my $host (@{$hostgroup->getMemberHosts})
	{
		# create an instance
		my $ci = main::createNagiosObject($cimClass);

		# fill in the remaining key properties
		$ci->setStringProperty('Name', $host->name);

		# fill in non-key properties
		$ci->setStringProperty('ElementName', $host->getProperty('alias'));
		$ci->setStringProperty('Address', $host->getProperty('address'));

		# fill in status elements
		Host::addHostStatus($ci, $host);

		# handle
		print "Returning host ", $host->name, "\n";
		$result->handle($ci);
	}
}

sub associators_listContainers
{
	my ($result, $memberName, $cimClass) = @_;

	my $nagios = main::getNagios();

	# iterate through each host group, then the members of that group
	# if the members list contains the host, add it to the list
	foreach my $hostgroup (@{$nagios->getAllHostGroups})
	{
		foreach my $host (@{$hostgroup->getMemberHosts})
		{
			if ($host->name eq $memberName)
			{
				# create an instance
				my $ci = main::createNagiosObject($cimClass);

				# fill in the remaining key properties
				$ci->setStringProperty('HostGroupName', $hostgroup->name);

				# fill in non-key properties
				$ci->setStringProperty('ElementName',
					$hostgroup->getProperty('alias'));

				# return it
				$result->handle($ci);

				# finished with this host group
				last;
			}
		}
	}
}

#######################################################################
#######################################################################

package HostService;
use base 'Config4GNU::WBEM::Provider';

sub associators
{
	my $providerclass = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole, $includeQualifiers, $includeClassOrigin,
		$propertyList) = @_;

	my $objectClass = $objectName->getClassName;
	print "ns = $ns\n";
	print "objectClass = $objectClass\n";

	# which side of the association is provided?
	if (lc($objectClass) eq 'config4gnu_nagiosmonitoredhost')
	{
		my $name = $objectName->getStringKeyValue('Name');
		my $cimClass = $env->getCIMOMHandle()->getClass($ns,
				'Config4GNU_NagiosMonitoredService');
		associators_listServices($result, $name, $cimClass);
	}
	else
	{
		die "Invalid classname $objectClass\n";
	}
}

sub associators_listServices
{
	my ($result, $name, $cimClass) = @_;

	my $nagios = main::getNagios();

	# find the specified host group
	my $host = $nagios->getHost($name);
	if (!defined($host))
	{
		# TODO: throw exception
		die "Host $name not found\n";
	}

	foreach my $service ($host->getServices)
	{
		$result->handle(Service::inst($service, $cimClass));
	}
}

package Host;
use base 'Config4GNU::WBEM::Provider';

sub addHostStatus
{
	my ($ci, $host) = @_;

	$ci->setStringProperty(
			'Status', $host->getStat('status'));
	$ci->setIntegerProperty(
			'TimeUp', $host->getStat('time_up'));
	$ci->setIntegerProperty(
			'TimeDown', $host->getStat('time_down'));
	$ci->setIntegerProperty(
			'TimeUnreachable', $host->getStat('time_unreachable'));
	$ci->setStringProperty(
			'LastUpdate', main::time_pl2cim($host->getStat('last_update')));
	$ci->setBooleanProperty(
			'ProcessPerformanceData',
			$host->getStat('process_performance_data'));
	$ci->setStringProperty(
			'PluginOutput', $host->getStat('plugin_output'));
	$ci->setBooleanProperty(
			'Acknowledged', $host->getStat('acknowledged'));
	$ci->setStringProperty(
			'LastNotificationTime',
			main::time_pl2cim($host->getStat('last_notification_time')));
	$ci->setIntegerProperty(
			'CurrentNotificationNumber',
			$host->getStat('current_notification_number'));
	$ci->setIntegerProperty(
			'ScheduledDowntimeDepth',
			$host->getStat('scheduled_downtime_depth'));
	$ci->setBooleanProperty(
			'NotificationsEnabled', $host->getStat('notifications_enabled'));
	$ci->setBooleanProperty(
			'EventHandlersEnabled', $host->getStat('event_handlers_enabled'));
	$ci->setBooleanProperty(
			'FailurePredictionEnabled',
			$host->getStat('failure_prediction_enabled'));
	$ci->setBooleanProperty(
			'ElementIsFlapping', $host->getStat('host_is_flapping'));
	$ci->setBooleanProperty(
			'ChecksEnabled', $host->getStat('checks_enabled'));
	$ci->setBooleanProperty(
			'FlapDetectionEnabled', $host->getStat('flap_detection_enabled'));
	$ci->setStringProperty(
			'LastCheckTime',
			main::time_pl2cim($host->getStat('last_check_time')));
	$ci->setStringProperty(
			'LastStateChange',
			main::time_pl2cim($host->getStat('last_state_change')));
	$ci->setStringProperty(
			'PercentStateChange', $host->getStat('percent_state_change'));
}

sub enumInstances
{
	my $self = shift;
	my ($env, $ns, $className, $localOnly, $deep, $includeQualifiers,
	  $includeClassOrigin, $result, $cimClass, $requestedClass) = @_;

	my $nagios = main::getNagios();

	foreach my $host (@{$nagios->{hosts}})
	{
		# create an instance
		my $ci = main::createNagiosObject($cimClass);

		# fill in the remaining key properties
		$ci->setStringProperty(
				"Name", $host->name);

		# fill in non-key properties
		$ci->setStringProperty(
				"ElementName", $host->getProperty('alias'));
		$ci->setStringProperty(
				"Address", $host->getProperty('address'));

		# fill in status elements
		addHostStatus($ci, $host);

		$result->handle($ci);
	}
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $nagios = main::getNagios();

	# find the specified host
	my $targetName = $instanceName->getStringKeyValue('Name');
	unless (defined $targetName)
	{
		warn "CIMObjectPath is missing key value 'Name'";
		return $cimClass->newInstance;
	}

	my $host = $nagios->getHost($targetName);
	if (!defined($host))
	{
		return undef;
	}

	# create a new instance and initialize with Nagios properties
	my $ci = main::createNagiosObject($cimClass);

	# fill in the remaining key properties
	$ci->setStringProperty(
			"Name", $host->name);

	# fill in non-key properties
	$ci->setStringProperty(
			"ElementName", $host->getProperty('alias'));
	$ci->setStringProperty(
			"Address", $host->getProperty('address'));

	# fill in status elements
	addHostStatus($ci, $host);

	return $ci;
}

package Service;
use base 'Config4GNU::WBEM::Provider';

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $hostname = $instanceName->getStringKeyValue('HostName');
	my $servicedesc = $instanceName->getStringKeyValue('ServiceDescription');

	my $nagios = main::getNagios();

	# find the specified host
	my $host = $nagios->getHost($hostname);
	if (!defined($host))
	{
		# TODO: throw exception
		die "Host $hostname not found\n";
	}

	# find the specified service
	my $service = $host->getService($servicedesc);
	if (!defined($service))
	{
		die "Service $servicedesc not found for host $hostname\n";
	}
	return inst($service, $cimClass);
}

sub inst
{
	my ($service, $cimClass) = @_;

	# create an instance
	my $ci = main::createNagiosObject($cimClass);

	# fill in the remaining key properties
	$ci->setStringProperty('HostName', $service->{HOSTNAME});
	$ci->setStringProperty('ServiceDescription',
			$service->{SERVICEDESCRIPTION});

	# fill in non-key properties
	my $hostalias = $service->{HOSTNAME};
	if (defined $service->{HOST})
	{
		my $host = $service->{HOST};
		$hostalias = $host->getProperty('alias');
		$ci->setStringProperty('HostAlias', $hostalias);
		$ci->setStringProperty('HostAddress', $host->getProperty('address'));
	}
	$ci->setStringProperty('ElementName',
		"$hostalias/" . $service->{SERVICEDESCRIPTION});

	# fill in status elements
	$ci->setStringProperty(
			'Status', $service->getStat('status'));
	my ($retryNumber, $maxRetries) = split(/\//,
			$service->getStat('retry_number'));
	$ci->setIntegerProperty('RetryNumber', $retryNumber);
	$ci->setIntegerProperty('MaxRetries', $maxRetries);
	$ci->setStringProperty(
			'StateType', $service->getStat('state_type'));
	$ci->setStringProperty(
			'CheckType', $service->getStat('check_type'));
	$ci->setStringProperty(
			'LastHardState', $service->getStat('last_hard_state'));
	$ci->setIntegerProperty(
			'TimeOK', $service->getStat('time_ok'));
	$ci->setIntegerProperty(
			'TimeUnknown', $service->getStat('time_unknown'));
	$ci->setIntegerProperty(
			'TimeWarning', $service->getStat('time_warning'));
	$ci->setIntegerProperty(
			'TimeCritical', $service->getStat('time_critical'));
	$ci->setStringProperty(
			'LastUpdate', main::time_pl2cim($service->getStat('last_update')));
	$ci->setBooleanProperty(
			'AcceptPassiveChecks',
			$service->getStat('accept_passive_checks'));
	$ci->setBooleanProperty(
			'ProcessPerformanceData',
			$service->getStat('process_performance_data'));
	$ci->setStringProperty(
			'PluginOutput', $service->getStat('plugin_output'));
	$ci->setBooleanProperty(
			'Acknowledged', $service->getStat('problem_acknowledged'));
	$ci->setStringProperty(
			'LastNotificationTime',
			main::time_pl2cim($service->getStat('last_notification_time')));
	$ci->setIntegerProperty(
			'CurrentNotificationNumber',
			$service->getStat('current_notification_number'));
	$ci->setIntegerProperty(
			'ScheduledDowntimeDepth',
			$service->getStat('scheduled_downtime_depth'));
	$ci->setBooleanProperty(
			'NotificationsEnabled', $service->getStat('notifications_enabled'));
	$ci->setBooleanProperty(
			'EventHandlersEnabled',
			$service->getStat('event_handlers_enabled'));
	$ci->setBooleanProperty(
			'FailurePredictionEnabled',
			$service->getStat('failure_prediction_enabled'));
	$ci->setBooleanProperty(
			'ElementIsFlapping', $service->getStat('service_is_flapping'));
	$ci->setBooleanProperty(
			'ChecksEnabled', $service->getStat('checks_enabled'));
	$ci->setBooleanProperty(
			'FlapDetectionEnabled',
			$service->getStat('flap_detection_enabled'));
	$ci->setStringProperty(
			'LastCheckTime',
			main::time_pl2cim($service->getStat('last_check_time')));
	$ci->setStringProperty(
			'NextCheckTime',
			main::time_pl2cim($service->getStat('next_check_time')));
	$ci->setStringProperty(
			'LastStateChange',
			main::time_pl2cim($service->getStat('last_state_change')));
	$ci->setStringProperty(
			'PercentStateChange', $service->getStat('percent_state_change'));
	
	return $ci;
}
