#!/usr/bin/perl

# Copyright 2004 Jason Long

# This file is part of config4gnu-wbem.
#
# Config4gnu-wbem is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Config4gnu-wbem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Config4gnu-wbem; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# USAGE:
#   Import the associated schema file (Samba.mof) into the OpenWBEM
#   CIMOM using owmofc.
#   Install this file to $(OPENWBEM_PREFIX)/lib/openwbem/perlproviders/

use strict;
use warnings;
use Config4GNU::WBEM;
use Config4GNU::Samba;

# standard classes
use constant CLASSNAME_SERVICE => 'Config4GNU_SambaService';
use constant CLASSNAME_SHARE => 'Config4GNU_SambaShare';
use constant CLASSNAME_SHARESETTINGS => 'Config4GNU_SambaShareSettings';
use constant CLASSNAME_GLOBALSETTINGS => 'Config4GNU_SambaGlobalSettings';

# association classes
use constant CLASSNAME_HOSTEDSERVICE => 'Config4GNU_SambaHostedService';
use constant CLASSNAME_HOSTEDSHARE => 'Config4GNU_SambaHostedShare';
use constant CLASSNAME_SERVICEGLOBALSETTINGS
	=> 'Config4GNU_SambaServiceGlobalSettings';
use constant CLASSNAME_SERVICEDEFAULTSETTINGS
	=> "Config4GNU_SambaServiceDefaultSettings";
use constant CLASSNAME_SHARESHARESETTINGS
	=> 'Config4GNU_SambaShareShareSettings';

# initialize standard class providers
Service->register(CLASSNAME_SERVICE);
Share->register(CLASSNAME_SHARE);
ShareSettings->register(CLASSNAME_SHARESETTINGS);
GlobalSettings->register(CLASSNAME_GLOBALSETTINGS);

# initialize association class providers
HostedService->register(CLASSNAME_HOSTEDSERVICE);
HostedShare->register(CLASSNAME_HOSTEDSHARE);
ServiceGlobalSettings->register(CLASSNAME_SERVICEGLOBALSETTINGS);
ServiceGlobalSettings->register(CLASSNAME_SERVICEDEFAULTSETTINGS);
ShareShareSettings->register(CLASSNAME_SHARESHARESETTINGS);

sub getSamba
{
	my $configfile = shift;
	my $samba = Config4GNU::Samba->new($configfile);
	return $samba;
}

sub getLocalSystemName
{
	my $hostname = `hostname`;
	chomp $hostname;
	return $hostname;
}

sub setCIMInstanceProperty
{
	my ($ci, $key, $type, $propvalue) = @_;

	if (defined $propvalue)
	{
		if ($type eq 'string')
		{
			$ci->setStringProperty($key, $propvalue);
		}
		elsif ($type eq 'boolean')
		{
			$ci->setBooleanProperty($key, $propvalue eq 'true');
		}
		elsif ($type eq 'int')
		{
			$ci->setIntegerProperty($key, $propvalue);
		}
		else
		{
			warn "Unknown type $type";
		}
	}
}

sub modifySambaSection
{
	my ($section, $modifiedInstance, $previousInstance, $mapping) = @_;

	foreach my $key (keys %{$mapping})
	{
		my ($type, $smb_name, $default) = @{$mapping->{$key}};
		
		if ($type eq 'string')
		{
			my $oldvalue = $previousInstance->getStringProperty($key);
			my $newvalue = $modifiedInstance->getStringProperty($key);
			
			if ($oldvalue ne $newvalue)
			{
				print "$smb_name => '$newvalue' (string) (from $oldvalue)\n";
				$section->setPropertyValue($smb_name, $newvalue);
			}
		}
		elsif ($type eq 'boolean')
		{
			my $oldvalue = $previousInstance->getBooleanProperty($key);
			my $newvalue = $modifiedInstance->getBooleanProperty($key);
			
			if ($oldvalue != $newvalue)
			{
				my $newvalueasstring = $newvalue ? 'true' : 'false';
				print "$smb_name => '$newvalueasstring' (boolean)\n";
				$section->setPropertyValue($smb_name, $newvalueasstring);
			}
		}
		elsif ($type eq 'int')
		{
			my $oldvalue = $previousInstance->getIntegerProperty($key);
			my $newvalue = $modifiedInstance->getIntegerProperty($key);
			
			if ($oldvalue != $newvalue)
			{
				print "$smb_name => $newvalue (int) (from $oldvalue)\n";
				$section->setPropertyValue($smb_name, $newvalue);
			}
		}
		else
		{
			warn "Unknown type $type";
		}
	}
}

#######################################################################
##                                                                   ##
##                            SERVICE PROVIDER                       ##
##                                                                   ##
#######################################################################

package Service;
use base 'Config4GNU::WBEM::SimpleProvider';
use constant CLASSNAME_SERVICE => main::CLASSNAME_SERVICE;

@Service::CONFIG_FILES = ('/etc/samba/smb.conf');

sub enumInstanceNames
{
	my $providerclass = shift;
	my ($env, $ns, $className, $result, $cimClass) = @_;

	my @config_files = grep(-f $_, @Service::CONFIG_FILES);
	foreach my $config_file (@config_files)
	{
		my $mcop = owperl::CIMObjectPath->new(CLASSNAME_SERVICE, $ns);
		$mcop->addStringKeyValue('Name', $config_file);
		$mcop->addStringKeyValue('CreationClassName', CLASSNAME_SERVICE);
		my $hostname = main::getLocalSystemName;
		$mcop->addStringKeyValue('SystemName', $hostname);
		$mcop->addStringKeyValue('SystemCreationClassName', 'CIM_System');
		$result->handle($mcop);
	}
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $config_file = $instanceName->getStringKeyValue('Name');
	if (not -f $config_file)
	{
		die "Invalid Samba service name $config_file\n";
	}

	# create a service instance
	my $ci = $cimClass->newInstance;

	# fill in the key properties
	$ci->setStringProperty('Name', $config_file);
	$ci->setStringProperty('CreationClassName', CLASSNAME_SERVICE);
	$ci->setStringProperty('SystemName',
		$instanceName->getStringKeyValue('SystemName'));
	$ci->setStringProperty('SystemCreationClassName',
		$instanceName->getStringKeyValue('SystemCreationClassName'));

	# fill in other properties
	$ci->setStringProperty('ConfigFile', $config_file);
	$ci->setStringProperty('ElementName', "Samba");

	return $ci;
}

#######################################################################
#######################################################################

package HostedService;
use base 'Config4GNU::WBEM::SimpleAssociationProvider';
use constant CLASSNAME_SERVICE => main::CLASSNAME_SERVICE;

sub getKeyInfo
{
	return (
		'ref', 'Antecedent',
		'ref', 'Dependent'
	);
}

sub referenceNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $resultClass) = @_;

	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, CLASSNAME_SERVICE))
	{
		my $mcop = owperl::CIMObjectPath->new(
			$objectName->getStringKeyValue('SystemCreationClassName'), $ns);
		$mcop->setStringKeyValue('Name',
			$objectName->getStringKeyValue('SystemName'));
		$mcop->setStringKeyValue('CreationClassName',
			$objectName->getStringKeyValue('SystemCreationClassName'));
		
		my $cop = owperl::CIMObjectPath->new(main::CLASSNAME_HOSTEDSERVICE, $ns);
		$cop->setRefKeyValue('Antecedent', $mcop);
		$cop->setRefKeyValue('Dependent', $objectName);
		$result->handle($cop);
	}
	elsif ($provider->is_subclass($env, $ns, $baseClass, 'CIM_System'))
	{
		my $systemName = $objectName->getStringKeyValue('Name');
		if ($systemName eq main::getLocalSystemName)
		{
			my $config_file = $Service::CONFIG_FILES[0];
			my $mcop = owperl::CIMObjectPath->new(CLASSNAME_SERVICE, $ns);
			$mcop->setStringKeyValue('Name', $config_file);
			$mcop->setStringKeyValue('CreationClassName', CLASSNAME_SERVICE);
			$mcop->setStringKeyValue('SystemName', $systemName);
			$mcop->setStringKeyValue('SystemCreationClassName',
				$objectName->getStringKeyValue('CreationClassName'));
		
			my $cop = owperl::CIMObjectPath->new(main::CLASSNAME_HOSTEDSERVICE, $ns);
			$cop->setRefKeyValue('Antecedent', $objectName);
			$cop->setRefKeyValue('Dependent', $mcop);		
			$result->handle($cop);
		}
	}
}

#######################################################################
#######################################################################

package HostedShare;
use base 'Config4GNU::WBEM::SimpleAssociationProvider';

sub getKeyInfo
{
	return (
		'ref', 'Antecedent',
		'ref', 'Dependent'
	);
}

sub referenceNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $resultClass) = @_;

	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_SERVICE))
	{
		my $configfile = $objectName->getStringKeyValue('Name');
		
		# because the Share class does not have any way to identify what
		# instance of the Samba service it belongs to, I can only list
		# shares for the default configuration file.
		unless ($configfile eq $Service::CONFIG_FILES[0])
		{
			return;
		}

		my $samba = Config4GNU::Samba->new($configfile);
		foreach my $section (@{$samba->children})
		{
			next if ($section->name eq 'global');
			
			my $mcop = owperl::CIMObjectPath->new(main::CLASSNAME_SHARE, $ns);
			$mcop->addStringKeyValue('Name', $section->name);
			$mcop->addStringKeyValue('CreationClassName',
				main::CLASSNAME_SHARE);
			$mcop->addStringKeyValue('SystemName',
				$objectName->getStringKeyValue('SystemName'));
			$mcop->addStringKeyValue('SystemCreationClassName',
				$objectName->getStringKeyValue('SystemCreationClassName'));
			
			my $cop = owperl::CIMObjectPath->new(main::CLASSNAME_HOSTEDSHARE, $ns);
			$cop->setRefKeyValue('Antecedent', $objectName);
			$cop->setRefKeyValue('Dependent', $mcop);
			$result->handle($cop);
		}
	}
	elsif ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_SHARE))
	{
		# currently our scheme only allows for share objects from
		# the "default" config file... that is, the first config file
		# listed in the CONFIG_FILES array
		my $configfile = $Service::CONFIG_FILES[0];
		
		my $mcop = owperl::CIMObjectPath->new(main::CLASSNAME_SERVICE, $ns);
		$mcop->addStringKeyValue('Name', $configfile);
		$mcop->addStringKeyValue('CreationClassName', main::CLASSNAME_SERVICE);
		$mcop->addStringKeyValue('SystemName',
			$objectName->getStringKeyValue('SystemName'));
		$mcop->addStringKeyValue('SystemCreationClassName',
			$objectName->getStringKeyValue('SystemCreationClassName'));
		
		my $cop = owperl::CIMObjectPath->new(main::CLASSNAME_HOSTEDSHARE, $ns);
		$cop->setRefKeyValue('Antecedent', $mcop);
		$cop->setRefKeyValue('Dependent', $objectName);
		$result->handle($cop);
	}
}

#######################################################################
##                                                                   ##
##                            SHARE PROVIDER                         ##
##                                                                   ##
#######################################################################

package Share;
use base 'Config4GNU::WBEM::SimpleProvider';
use constant CLASSNAME_SHARE => main::CLASSNAME_SHARE;

sub enumInstanceNames
{
	my $self = shift;
	my ($env, $ns, $className, $result, $cimClass) = @_;

	my $samba = main::getSamba();

	foreach my $share (@{$samba->children})
	{
		next if ($share->name eq 'global');

		my $mcop = owperl::CIMObjectPath->new(CLASSNAME_SHARE, $ns);
		$mcop->addStringKeyValue('Name', $share->name);
		$mcop->addStringKeyValue('CreationClassName', CLASSNAME_SHARE);
		my $hostname = main::getLocalSystemName;
		$mcop->addStringKeyValue('SystemName', $hostname);
		$mcop->addStringKeyValue('SystemCreationClassName', 'CIM_System');

		$result->handle($mcop);
	}
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $targetName = $instanceName->getStringKeyValue('Name');
	my $samba = main::getSamba();

	my $share = find_section($samba, $targetName);
	if (not defined $share)
	{
		die "Share not found.\n";
	}

	# create an instance
	my $ci = $cimClass->newInstance;

	# fill in the key properties
	$ci->setStringProperty('Name', $share->name);
	$ci->setStringProperty('CreationClassName', CLASSNAME_SHARE);
	$ci->setStringProperty('SystemName',
		$instanceName->getStringKeyValue('SystemName'));
	$ci->setStringProperty('SystemCreationClassName',
		$instanceName->getStringKeyValue('SystemCreationClassName'));

	# fill in other properties
	$ci->setBooleanProperty('SharingDirectory', 1);
	$ci->setStringProperty('ElementName', $share->name);

	my $propvalue;
	$propvalue = $share->getPropertyValue('path');
	if (defined $propvalue)
	{
		$ci->setStringProperty('Path', $propvalue);
	}
	$propvalue = $share->getPropertyValue('comment');
	if (defined $propvalue)
	{
		$ci->setStringProperty('Comment', $propvalue);
	}
	$propvalue = $share->getPropertyValue('read only');
	if (defined $propvalue)
	{
		$ci->setBooleanProperty('ReadOnly', $propvalue eq 'true');
	}
	$propvalue = $share->getPropertyValue('guest ok');
	if (defined $propvalue)
	{
		$ci->setBooleanProperty('GuestOK', $propvalue eq 'true');
	}
	$propvalue = $share->getPropertyValue('browseable');
	if (defined $propvalue)
	{
		$ci->setBooleanProperty('Browseable', $propvalue eq 'true');
	}

	return $ci;
}

sub find_section
{
	my ($samba, $targetName) = @_;
	foreach my $share (@{$samba->children})
	{
		return $share if ($share->name eq $targetName);
	}
	return undef;
}

sub createInstance
{
	my $provider = shift;
	my ($env, $ns, $cimInstance) = @_;

	my $sectionName = $cimInstance->getStringProperty('Name');

	# create an empty section in the smb.conf file
	my $samba = main::getSamba();
	my $share = $samba->newSection($sectionName);
	
	# fill in some of the more common properties
	# path
	my $path = $cimInstance->getStringProperty('Path');
	if (defined $path)
	{
		$share->setPropertyValue('path', $path);
	}
	# read only
	my $writeable = $cimInstance->getBooleanProperty('Writeable');
	$share->setPropertyValue('read only', $writeable ? 'false' : 'true');
	# comment
	my $comment = $cimInstance->getStringProperty("Comment");
	if (defined $comment)
	{
		$share->setPropertyValue('comment', $comment);
	}
	
	# save the new smb.conf file
	$samba->save;

	# create a path for the new instance
	my $cop = owperl::CIMObjectPath->new(CLASSNAME_SHARE, $ns);
	$cop->addStringKeyValue('Name', $sectionName);
	$cop->addStringKeyValue('CreationClassName', CLASSNAME_SHARE);
	$cop->addStringKeyValue('SystemName',
		$cimInstance->getStringProperty('SystemName'));
	$cop->addStringKeyValue('SystemCreationClassName',
		$cimInstance->getStringProperty('SystemCreationClassName'));

	print "returning: ", $cop->toString, "\n";
	return $cop;
}

sub deleteInstance
{
	my $provider = shift;
	my ($env, $ns, $cop) = @_;

	my $sectionName = $cop->getStringKeyValue('Name');

	print "deleting section $sectionName\n";

	# remove the associated configuration section
	my $samba = main::getSamba();
	$samba->deleteSection($sectionName);

	print "saving modified smb.conf file\n";
	$samba->save;
}

#######################################################################
##                                                                   ##
##                       SHARE SETTINGS PROVIDER                     ##
##                                                                   ##
#######################################################################

package ShareSettings;
use base 'Config4GNU::WBEM::SimpleProvider';
use constant CLASSNAME_SHARESETTINGS => main::CLASSNAME_SHARESETTINGS;

		# CIM name            type       smb.conf name     default
		# --------            ----       -------------     -------
my %PROPERTY_MAPPING =
	(	'Description' =>    [ 'string',  'comment',        '' ],
		'AdminUsers' =>     [ 'string',  'admin users',    '' ],
		'Available' =>      [ 'boolean', 'available',      'true' ],
		'BlockSize' =>      [ 'int',     'block size',     1024 ],
		'BlockingLocks' =>  [ 'boolean', 'blocking locks', 'true' ],
		'Browseable' =>     [ 'boolean', 'browseable',     'true' ],
		'CaseSensitive' =>  [ 'boolean', 'case sensitive', 'false' ],
		'Copy' =>           [ 'string',  'copy',           undef ],
		'CreateMask' =>     [ 'string',  'create mask',    '0744' ],
		'CscPolicy' =>      [ 'string',  'csc policy',     'manual' ],
		'DefaultCase' =>    [ 'string',  'default case',   'lower' ],
		'DefaultDevMode' => [ 'boolean', 'default dev mode', 'false' ],
		'DeleteReadonly' => [ 'boolean', 'delete readonly', 'false' ],
		'DeleteVetoFiles' => [ 'boolean', 'delete veto files', 'false' ],
		'DirectoryMask' =>  [ 'string',  'directory mask', '0755' ],
		'DirectorySecurityMask' => [ 'string', 'directory security mask', '0777' ],
		'DontDescend' =>    [ 'string',  'dont descend',   undef ],
		'DosFilemode' =>    [ 'boolean', 'dos filemode',   'false' ],
		'DosFiletimeResolution' => [ 'boolean', 'dos filetime resolution', 'false' ],
		'DosFiletimes' =>   [ 'boolean', 'dos filetimes',  'false' ],
		'FakeDirectoryCreateTimes' => [ 'boolean', 'fake directory create times', 'false' ],
		'FakeOplocks' =>    [ 'boolean', 'fake oplocks', 'false' ],
		'FollowSymlinks' => [ 'boolean', 'follow symlinks', 'true' ],
		'ForceCreateMode' => [ 'string', 'force create mode', '000' ],
		'ForceDirectoryMode' => [ 'string', 'force directory mode', '000' ],
		'ForceDirectorySecurityMode' => [ 'string', 'force directory security mode', '0' ],
		'ForceGroup' =>    [ 'string',   'force group',   undef ],
		'ForceSecurityMode' => [ 'string', 'force security mode', '0' ],
		'ForceUnknownAclUser' => [ 'boolean', 'force unknown acl user', 'false' ],
		'ForceUser' =>     [ 'string',  'force user',     undef ],
		'FsType' =>        [ 'string',  'fs type',        'NTFS' ],
		'GuestAccount' =>  [ 'string',  'guest account',  'nobody' ],
		'GuestOK' =>       [ 'boolean', 'guest ok',       'false' ],
		'GuestOnly' =>     [ 'boolean', 'guest only',     'false' ],
		'HideDotFiles' =>  [ 'boolean', 'hide dot files', 'true' ],
		'HideFiles' =>     [ 'string',  'hide files',     undef ],
		'HideUnreadable' => [ 'boolean', 'hide unreadable', 'false' ],
		'HostsAllow' =>    [ 'string',  'hosts allow',    undef ],
		'HostsDeny' =>     [ 'string',  'hosts deny',     undef ],
		'InheritAcls' =>   [ 'boolean', 'inherit acls',   'false' ],
		'InheritPermissions' => [ 'boolean', 'inherit permissions', 'false' ],
		'InvalidUsers' =>  [ 'string',  'invalid users',  undef ],
		'Level2Oplocks' => [ 'boolean', 'level2 oplocks', 'true' ],
		'Locking' =>       [ 'boolean', 'locking',        'true' ],
		'LppauseCommand' => [ 'string', 'lppause command', undef ],
		'LpqCommand' =>    [ 'string',  'lpq command',    undef ],
		'LpresumeCommand' => [ 'string', 'lpresume command', undef ],
		'LprmCommand' =>   [ 'string',  'lprm command',   undef ],
		'MagicOutput' =>   [ 'string',  'magic output',   undef ],
		'MagicScript' =>   [ 'string',  'magic script',   undef ],
		'MangleCase' =>    [ 'boolean', 'mangle case',    'false' ],
		'MangledMap' =>    [ 'string',  'mangled map',    undef ],
		'MangledNames' =>  [ 'boolean', 'mangled names',  'true' ],
		'ManglingChar' =>  [ 'string',  'mangling char',  '~' ],
		'MapArchive' =>    [ 'boolean', 'map archive',    'true' ],
		'MapHidden' =>     [ 'boolean', 'map hidden',     'false' ],
		'MapSystem' =>     [ 'boolean', 'map system',     'false' ],
		'MaxConnections' => [ 'int',    'max connections', 0 ],
		'MaxPrintJobs' =>  [ 'int',     'max print jobs',  1000 ],
		'MinPrintSpace' => [ 'int',     'min print space', 0 ],
		'MsdfsRoot' =>     [ 'boolean', 'msdfs root',     'false' ],
		'NtAclSupport' =>  [ 'boolean', 'nt acl support', 'true' ],
		'OnlyUser' =>      [ 'boolean', 'only user',      'false' ],
		'OplockContentionLimit' => [ 'int', 'oplock contention limit', 2 ],
		'Oplocks' =>       [ 'boolean', 'oplocks',        'true' ],
		'Path' =>          [ 'string',  'path',           undef ],
		'PosixLocking' =>  [ 'boolean', 'posix locking',  'true' ],
		'Postexec' =>      [ 'string',  'postexec',       undef ],
		'Postscript' =>    [ 'boolean', 'postscript',     'false' ],
		'Preexec' =>       [ 'string',  'preexec',        undef ],
		'PreexecClose' =>  [ 'boolean', 'preexec close',  'false' ],
		'PreserveCase' =>  [ 'boolean', 'preserve case',  'true' ],
		'PrintCommand' =>  [ 'string',  'print command',  undef ],
		'Printable' =>     [ 'boolean', 'printable',      'false' ],
		'PrinterAdmin' =>  [ 'string',  'printer admin',  '' ],
		'PrinterDriver' => [ 'string',  'printer driver', undef ],
		'PrinterDriverLocation' => [ 'string', 'printer driver location', undef ],
		'PrinterName' =>   [ 'string',  'printer name',   undef ],
		'Printing' =>      [ 'string',  'printing',       undef ],
		'ProfileAcls' =>   [ 'boolean', 'profile acls',   'false' ],
		'QueuepauseCommand' => [ 'string', 'queuepause command', undef ],
		'QueueresumeCommand' => [ 'string', 'queueresume command', undef ],
		'ReadList' =>      [ 'string',  'read list',      '' ],
		'ReadOnly' =>      [ 'boolean', 'read only',      'true' ],
		'RootPostexec' =>  [ 'string',  'root postexec',  undef ],
		'RootPreexec' =>   [ 'string',  'root preexec',   undef ],
		'RootPreexecClose' => [ 'boolean', 'root preexec close', 'false' ],
		'SecurityMask' =>  [ 'string',  'security mask',  '0777' ],
		'SetDirectory' =>  [ 'boolean', 'set directory',  'false' ],
		'ShareModes' =>    [ 'boolean', 'share modes',    'true' ],
		'ShortPreserveCase' => [ 'boolean', 'short preserve case', 'true' ],
		'StrictAllocate' => [ 'boolean', 'strict allocate', 'false' ],
		'StrictLocking' => [ 'boolean', 'strict locking', 'false' ],
		'StrictSync' =>    [ 'boolean', 'strict sync',    'false' ],
		'SyncAlways' =>    [ 'boolean', 'sync always',    'false' ],
		'UseClientDriver' => [ 'boolean', 'use client driver', 'false' ],
		'Username' =>      [ 'string',  'username',       '' ],
		'UseSendfile' =>   [ 'boolean', 'use sendfile',   'false' ],
		'ValidUsers' =>    [ 'string',  'valid users',    undef ],
		'VetoFiles' =>     [ 'string',  'veto files',     undef ],
		'VetoOplockFiles' => [ 'string', 'veto oplock files', undef ],
		'VfsOptions' =>    [ 'string',  'vfs options',    undef ],
		'Volume' =>        [ 'string',  'volume',         undef ],
		'WideLinks' =>     [ 'boolean', 'wide links',     'true' ],
		'WriteCacheSize' => [ 'int',    'write cache size', 0 ],
		'WriteList' =>     [ 'string',  'write list',     '' ]
	);

sub enumInstanceNames
{
	my $provider = shift;
	my ($env, $ns, $className, $result, $cimClass) = @_;

	foreach my $configfile (grep(-f $_, @Service::CONFIG_FILES))
	{
		my $samba = main::getSamba($configfile);
		foreach my $share (@{$samba->children})
		{
			next if ($share->name eq 'global');

			my $mcop = owperl::CIMObjectPath->new(CLASSNAME_SHARESETTINGS, $ns);
			$mcop->addStringKeyValue('Name', $share->name);
			$mcop->addStringKeyValue('ConfigFile', $configfile);

			$result->handle($mcop);
		}
	}
}

sub getInstance
{
	my $provider = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $configfile = $instanceName->getStringKeyValue('ConfigFile');
	my $shareName = $instanceName->getStringKeyValue('Name');
	unless (defined($configfile) && defined($shareName))
	{
		die "CIMObjectPath missing key value(s)\n";
	}
	
	my $samba = main::getSamba($configfile);
	my $share = $samba->getSection($shareName);
	if (not defined $share)
	{
		die "Share not found.\n";
	}

	# create an instance
	my $ci = $cimClass->newInstance;

	# fill in the key properties
	$ci->setStringProperty('ConfigFile', $samba->{'FILENAME'});
	$ci->setStringProperty('Name', $share->name);

	# fill in other properties
	$ci->setStringProperty('ElementName',
		$share->name eq 'global' ? 'Default Share Settings'
		: $share->name);

	# samba allows sections to "copy" attributes from other sections,
	# and implicitly from the global section.
	my @templates = ($share);
	my $src;
	while (defined ($src = $templates[$#templates]->getPropertyValue('copy')))
	{
		my $src_section = $samba->getSection($src);
		if (defined $src_section)
		{
			push @templates, $src_section;
		}
	}
	my $global_section = $samba->getSection('global');
	if (defined $global_section && $global_section != $share)
	{
		push @templates, $global_section;
	}

	foreach my $key (keys %PROPERTY_MAPPING)
	{
		my $type = $PROPERTY_MAPPING{$key}->[0];
		my $smb_name = $PROPERTY_MAPPING{$key}->[1];
		my $default = $PROPERTY_MAPPING{$key}->[2];
		
		my $propvalue;
		foreach my $section (@templates)
		{
			$propvalue = $section->getPropertyValue($smb_name);
			last if (defined $propvalue);
		}
		if (!defined($propvalue))
		{
			# setting is not defined, so use default (if present)
			$propvalue = $default;
		}
		main::setCIMInstanceProperty($ci, $key, $type, $propvalue);
	}
	return $ci;
}

sub modifyInstance
{
	my $provider = shift;
	my ($env, $ns, $modifiedInstance, $previousInstance,
		$includeQualifiers, $propertyList, $theClass) = @_;

	# key properties
	my $ci = $modifiedInstance;
	my $config_file = $ci->getStringProperty('ConfigFile');
	my $share_name = $ci->getStringProperty('Name');
	
	unless (defined($config_file) && defined($share_name))
	{
		die "Missing key properties\n";
	}
	
	my $samba = main::getSamba($config_file);
	my $share = $samba->getSection($share_name);
	unless (defined $share)
	{
		die "Invalid share name\n";
	}

	main::modifySambaSection($share, $modifiedInstance, $previousInstance,
			\%PROPERTY_MAPPING);

	print "Saving modified Samba configuration file\n";
	$samba->save;
}

#######################################################################
#######################################################################

package GlobalSettings;
use base 'Config4GNU::WBEM::SimpleProvider';

		# CIM name              type       smb.conf name     default
		# --------              ----       -------------     -------
my %GLOBAL_PROPERTY_MAPPING =
	(	'AclCompatibility' => [ 'string',  'acl compatibility', undef ],
		'AddPrinterCommand' => [ 'string', 'add printer command', undef ],
		'AddShareCommand' =>  [ 'string',  'add share command', undef ],
		'AddUserScript' =>    [ 'string',  'add user script', undef ],
		'AllowTrustedDomains' => [ 'boolean', 'allow trusted domains', 'false' ],
		'AnnounceAs' =>       [ 'string',  'announce as',    'NT' ],
		'AnnounceVersion' =>  [ 'string',  'announce version', '4.9' ],
		'BindInterfacesOnly' => [ 'boolean', 'bind interfaces only', 'false' ],
		'BrowseList' =>       [ 'boolean', 'browse list',    'true' ],
		'ChangeNotifyTimeout' => [ 'int',  'change notify timeout', 60 ],
		'ChangeShareCommand' => [ 'string', 'change share command', undef ],
		'CharacterSet' =>     [ 'string',  'character set',  undef ],
		'ClientCodePage' =>   [ 'string',  'client code page', '850' ],
		'CodePageDirectory' => [ 'string', 'code page directory', undef ],
		'CodingSystem' =>     [ 'string',  'coding system',  undef ],
		'Deadtime' =>         [ 'int',     'deadtime',       0 ],
		'DebugHiresTimestamp' => [ 'boolean', 'debug hires timestamp', 'false' ],
		'DebugPid' =>         [ 'boolean', 'debug pid',      'false' ],
		'DebugTimestamp' =>   [ 'boolean', 'debug timestamp', 'true' ],
		'DebugUid' =>         [ 'boolean', 'debug uid',      'false' ],
		'DefaultService' =>   [ 'string',  'default service', undef ],
		'DeletePrinterCommand' => [ 'string','delete printer command', undef ],
		'DeleteShareCommand' => [ 'string', 'delete share command', undef ],
		'DeleteUserScript' => [ 'string',  'delete user script', undef ],
		'DfreeCommand' =>     [ 'string',  'dfree command',  undef ],
		'DisableSpoolss' =>    [ 'boolean', 'disable spools',  'false' ],
		'DnsProxy' =>         [ 'boolean', 'dns proxy',      'true' ],
		'DomainAdminGroup' => [ 'string',  'domain admin group', undef ],
		'DomainGuestGroup' => [ 'string',  'domain guest group', undef ],
		'DomainLogons' =>     [ 'boolean', 'domain logons',  'false' ],
		'DomainMaster' =>     [ 'boolean', 'domain master',  undef ],
		'EncryptPasswords' => [ 'boolean', 'encrypt passwords', 'false' ],
		'EnhancedBrowsing' => [ 'boolean', 'enhanced browsing', 'true' ],
		'EnumportsCommand' => [ 'string',  'enumports command', undef ],
		'GetwdCache' =>       [ 'boolean', 'getwd cache',    'true' ],
		'HideLocalUsers' =>   [ 'boolean', 'hide local users', 'false' ],
		'HomedirMap' =>       [ 'boolean', 'homedir map',    '' ],
		'HostMsdfs' =>        [ 'boolean', 'host msdfs',     'false' ],
		'HostsEquiv' =>       [ 'string',  'hosts equiv',    undef ],
		'Interfaces' =>       [ 'string',  'interfaces',     undef ],
		'Keepalive' =>        [ 'int',     'keepalive',      300 ],
		'KernelOplocks' =>    [ 'boolean', 'kernel oplocks', 'true' ],
		'LanmanAuth' =>       [ 'boolean', 'lanman auth',    'true' ],
		'LargeReadwrite' =>   [ 'boolean', 'large readwrite', 'true' ],
		'LdapAdminDn' =>      [ 'string',  'ldap admin dn',  undef ],
		'LdapFilter' =>       [ 'string',  'ldap filter',    '(&(uid=%u)(objectclass=sambaAccount))' ],
		'LdapPort' =>         [ 'int',     'ldap port',      undef ],
		'LdapServer' =>       [ 'string',  'ldap server',    'localhost' ],
		'LdapSsl' =>          [ 'string',  'ldap ssl',       'on' ],
		'LdapSuffix' =>       [ 'string',  'ldap suffix',    undef ],
		'LmAnnounce' =>       [ 'string',  'lm announce',    'auto' ],
		'LmInterval' =>       [ 'int',     'lm interval',    60 ],
		'LoadPrinters' =>     [ 'boolean', 'load printers',  'true' ],
		'LocalMaster' =>      [ 'boolean', 'local master',   'true' ],
		'LockDirectory' =>    [ 'string',  'lock directory', undef ],
		'LockSpinCount' =>    [ 'int',     'lock spin count', 2 ],
		'LockSpinTime' =>     [ 'int',     'lock spin time', 10 ],
		'LogFile' =>          [ 'string',  'log file',       undef ],
		'LogLevel' =>         [ 'int',     'log level',      3 ],
		'LogonDrive' =>       [ 'string',  'logon drive',    'z:' ],
		'LogonHome' =>        [ 'string',  'logon home',     '\\\\%N\\%U' ],
		'LogonPath' =>        [ 'string',  'logon path',     '\\\\%N\\%U\\profile' ],
		'LogonScript' =>      [ 'string',  'logon script',   undef ],
		'LpqCacheTime' =>     [ 'int',     'lpq cache time', 10 ],
		'MachinePasswordTimeout' => [ 'int', 'machine password timeout', 604800 ],
		'MangledStack' =>     [ 'int',     'mangled stack',  50 ],
		'ManglingMethod' =>   [ 'string',  'mangling method', 'hash' ],
		'MapToGuest' =>       [ 'string',  'map to guest',   'Never' ],
		'MaxDiskSize' =>      [ 'int',     'max disk size',  0 ],
		'MaxLogSize' =>       [ 'int',     'max log size',   5000 ],
		'MaxMux' =>           [ 'int',     'max mux',        50 ],
		'MaxOpenFiles' =>     [ 'int',     'max open files', 10000 ],
		'MaxProtocol' =>      [ 'string',  'max protocol',   'NT1' ],
		'MaxSmbdProcesses' => [ 'int',     'max smbd processes', 0 ],
		'MaxTtl' =>           [ 'int',     'max ttl',        259200 ],
		'MaxWinsTtl' =>       [ 'int',     'max wins ttl',   518400 ],
		'MaxXmit' =>          [ 'int',     'max xmit',       16644 ],
		'MessageCommand' =>   [ 'string',  'message command', undef ],
		'MinPasswordLength' => [ 'int',    'min password length', 5 ],
		'MinProtocol' =>      [ 'string',  'min protocol',   'CORE' ],
		'MinWinsTtl' =>       [ 'int',     'min wins ttl',   21600 ],
		'NameResolveOrder' => [ 'string',  'name resolve order', 'lmhosts host wins bcast' ],
		'NetBIOSAliases' =>   [ 'string',  'netbios aliases', '' ],
		'NetBIOSName' =>      [ 'string',  'netbios name',   undef ],
		'NetBIOSScope' =>     [ 'string',  'netbios scope',  undef ],
		'NisHomedir' =>       [ 'boolean', 'nis homedir',    'false' ],
		'NtPipeSupport' =>    [ 'boolean', 'nt pipe support', 'true' ],
		'NtSmbSupport' =>     [ 'boolean', 'nt smb support', 'true' ],
		'NtStatusSupport' =>  [ 'boolean', 'nt status support', 'true' ],
		'NullPasswords' =>    [ 'boolean', 'null passwords', 'false' ],
		'ObeyPamRestrictions' => [ 'boolean', 'obey pam restrictions', 'false' ],
		'OplockBreakWaitTime' => [ 'int',  'oplock break wait time', 0 ],
		'OSLevel' =>          [ 'int',     'os level',       20 ],
		'Os2DriverMap' =>     [ 'string',  'os2 driver map', '' ],
		'PamPasswordChange' => [ 'boolean','pam password change', 'false' ],
		'PanicAction' =>      [ 'string',  'panic action',   '' ],
		'PasswdChat' =>       [ 'string',  'passwd chat',    undef ],
		'PasswdChatDebug' =>  [ 'boolean', 'passwd chat debug', 'false' ],
		'PasswdProgram' =>    [ 'string',  'passwd program', '/bin/passwd' ],
		'PasswordLevel' =>    [ 'int',     'password level', 0 ],
		'PasswordServer' =>   [ 'string',  'password server', '' ],
		'PidDirectory' =>     [ 'string',  'pid directory',  undef ],
		'PreferredMaster' =>  [ 'boolean', 'preferred master', undef ],
		'Preload' =>          [ 'string',  'preload',        undef ],
		'PrintcapName' =>     [ 'string',  'printcap name',  '/etc/printcap' ],
		'PrinterDriverFile' => [ 'string', 'printer driver file', undef ],
		'ReadBmpx' =>         [ 'boolean', 'read bmpx',      'false' ],
		'ReadRaw' =>          [ 'boolean', 'read raw',       'true' ],
		'ReadSize' =>         [ 'int',     'read size',      16384 ],
		'RemoteAnnounce' =>   [ 'string',  'remote announce', '' ],
		'RemoteBrowseSync' => [ 'string',  'remote browse sync', '' ],
		'RestrictAnonymous' => [ 'boolean','restrict anonymous', 'false' ],
		'RootDirectory' =>    [ 'string',  'root directory', '/' ],
		'Security' =>         [ 'string',  'security',       'user' ],
		'ServerString' =>     [ 'string',  'server string',  'Samba %v' ],
		'ShowAddPrinterWizard' => [ 'boolean', 'show add printer wizard', 'true' ],
		'SmbPasswdFile' =>    [ 'string',  'smb passwd file', undef ],
		'SocketAddress' =>    [ 'string',  'socket address', undef ],
		'SocketOptions' =>    [ 'string',  'socket options', 'TCP_NODELAY' ],
		'SourceEnvironment' => [ 'string', 'source environment', undef ],
		'Ssl' =>              [ 'boolean', 'ssl',            'false' ],
		'SslCACertDir' =>     [ 'string',  'ssl ca certdir', '/usr/local/ssl/certs' ],
		'SslCACertFile' =>    [ 'string',  'ssl ca certfile', '/usr/local/ssl/certs/trustedCAs.pem' ],
		'SslCiphers' =>       [ 'string',  'ssl ciphers',    undef ],
		'SslClientCert' =>    [ 'string',  'ssl client cert', '/usr/local/ssl/certs/smbclient.pem' ],
		'SslClientKey' =>     [ 'string',  'ssl client key', '/usr/local/ssl/private/smbclient.pem' ],
		'SslCompatibility' => [ 'boolean', 'ssl compatibility', 'false' ],
		'SslEgdSocket' =>     [ 'string',  'ssl egd socket', undef ],
		'SslEntropyBytes' =>  [ 'int',     'ssl entropy bytes', 255 ],
		'SslEntropyFile' =>   [ 'string',  'ssl entropy file', undef ],
		'SslHostsResign' =>   [ 'string',  'ssl hosts resign', '' ],
		'SslRequireClientcert' => [ 'boolean', 'ssl require clientcert', 'false' ],
		'SslRequireServercert' => [ 'boolean', 'ssl require servercert', 'false' ],
		'SslServerCert' =>    [ 'string',  'ssl server cert', '' ],
		'SslServerKey' =>     [ 'string',  'ssl server key', '' ],
		'SslVersion' =>       [ 'string',  'ssl version',    'ssl2or3' ],
		'StatCache' =>        [ 'boolean', 'stat cache',     'true' ],
		'StatCacheSize' =>    [ 'int',     'stat cache size', 50 ],
		'StripDot' =>         [ 'boolean', 'strip dot',      'false' ],
		'Syslog' =>           [ 'int',     'syslog',         1 ],
		'SyslogOnly' =>       [ 'boolean', 'syslog only',    'false' ],
		'TemplateHomedir' =>  [ 'string',  'template homedir', '/home/%D/%U' ],
		'TemplateShell' =>    [ 'string',  'template shell', '/bin/false' ],
		'TimeOffset' =>       [ 'int',     'time offset',    0 ],
		'TimeServer' =>       [ 'boolean', 'time server',    'false' ],
		'TotalPrintJobs' =>   [ 'int',     'total print jobs', 0 ],
		'UnixExtensions' =>   [ 'boolean', 'unix extensions', 'false' ],
		'UnixPasswordSync' => [ 'boolean', 'unix password sync', 'false' ],
		'UpdateEncrypted' =>  [ 'boolean', 'update encrypted', 'false' ],
		'UseMmap' =>          [ 'boolean', 'use mmap',       'true' ],
		'UseRhosts' =>        [ 'boolean', 'use rhosts',     'false' ],
		'UsernameLevel' =>    [ 'int',     'username level', 0 ],
		'UsernameMap' =>      [ 'string',  'username map',   undef ],
		'Utmp' =>             [ 'boolean', 'utmp',           'false' ],
		'UtmpDirectory' =>    [ 'string',  'utmp directory', undef ],
		'ValidChars' =>       [ 'string',  'valid chars',    undef ],
		'WinbindCacheTime' => [ 'int',     'winbind cache time', 15 ],
		'WinbindEnumUsers' => [ 'boolean', 'winbind enum users', 'true' ],
		'WinbindEnumGroups' => [ 'boolean', 'winbind enum groups', 'true' ],
		'WinbindGid' =>       [ 'string',  'winbind gid',    '' ],
		'WinbindSeparator' => [ 'string',  'winbind separator', '\\' ],
		'WinbindUid' =>       [ 'string',  'winbind uid',    '' ],
		'WinbindUseDefaultDomain' => [ 'boolean', 'winbind use default domain', 'false' ],
		'WinsHook' =>         [ 'string',  'wins hook',      undef ],
		'WinsProxy' =>        [ 'boolean', 'wins proxy',     'false' ],
		'WinsServer' =>       [ 'string',  'wins server',    undef ],
		'WinsSupport' =>      [ 'boolean', 'wins support',   'false' ],
		'Workgroup' =>        [ 'string',  'workgroup',      'WORKGROUP' ],
		'WriteRaw' =>         [ 'boolean', 'write raw',      'true' ]
	);

sub getInstance
{
	my $providerclass = shift;
	my ($env, $ns, $instanceName, $localOnly, $includeQualifiers,
	  $includeClassOrigin, $propertyList, $cimClass) = @_;

	my $configfile = $instanceName->getStringKeyValue('ConfigFile');
	unless (defined($configfile))
	{
		die "CIMObjectPath missing key value(s)\n";
	}
	unless (length($configfile) > 0)
	{
		die "ConfigFile key is empty string\n";
	}

	my $samba = main::getSamba($configfile);
	my $global_section = $samba->getSection('global');

	# create a new instance
	my $ci = $cimClass->newInstance;
	
	# set key properties
	$ci->setStringProperty('ConfigFile', $samba->{'FILENAME'});	

	# set other properties
	$ci->setStringProperty('ElementName', 'Global Settings');

	foreach my $key (keys %GLOBAL_PROPERTY_MAPPING)
	{
		my $type = $GLOBAL_PROPERTY_MAPPING{$key}->[0];
		my $smb_name = $GLOBAL_PROPERTY_MAPPING{$key}->[1];
		my $default = $GLOBAL_PROPERTY_MAPPING{$key}->[2];

		my $propvalue;
		if (defined $global_section)
		{
			$propvalue = $global_section->getPropertyValue($smb_name);
		}
		if (!defined($propvalue))
		{
			# add hard to determine default values here...
			if ($key eq 'NetBIOSName')
			{
				chomp ($default = uc(`hostname -s`));
			}

			# setting is not defined, so use default (if present)
			$propvalue = $default;
		}
		main::setCIMInstanceProperty($ci, $key, $type, $propvalue);
	}
	return $ci;
}

sub modifyInstance
{
	my $providerclass = shift;
	my ($env, $ns, $modifiedInstance, $previousInstance,
		$includeQualifiers, $propertyList, $theClass) = @_;

	# key properties
	my $ci = $modifiedInstance;
	my $config_file = $ci->getStringProperty('ConfigFile');
	
	unless (defined($config_file))
	{
		die "Missing key properties\n";
	}
	
	my $samba = main::getSamba($config_file);
	my $global_section = $samba->getSection('global');
	unless (defined $global_section)
	{
		die "No 'global' section!\n";
	}

	main::modifySambaSection($global_section, $modifiedInstance,
			$previousInstance, \%GLOBAL_PROPERTY_MAPPING);

	print "Saving modified Samba configuration file\n";
	$samba->save;
}

#######################################################################
##                                                                   ##
##                  SERVICE GLOBAL SETTINGS PROVIDER                 ##
##                                     (association class)           ##
##                                                                   ##
#######################################################################

package ServiceGlobalSettings;
use base 'Config4GNU::WBEM::SimpleAssociationProvider';

sub getKeyInfo
{
	return (
		'ref', 'Element',
		'ref', 'Setting'
	);
}

sub referenceNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $resultClass) = @_;

	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_SERVICE))
	{
		my $configfile = $objectName->getStringKeyValue('Name');
		my $mcop;
		if (lc($resultClass) eq lc(main::CLASSNAME_SERVICEGLOBALSETTINGS))
		{
			$mcop = owperl::CIMObjectPath->new(
					main::CLASSNAME_GLOBALSETTINGS, $ns);
			$mcop->addStringKeyValue('ConfigFile', $configfile);
		}
		elsif (lc($resultClass) eq lc(main::CLASSNAME_SERVICEDEFAULTSETTINGS))
		{
			$mcop = owperl::CIMObjectPath->new(
					main::CLASSNAME_SHARESETTINGS, $ns);
			$mcop->addStringKeyValue('ConfigFile', $configfile);
			$mcop->addStringKeyValue('Name', 'global');
		}
		else
		{
			return;
		}
		my $cop = owperl::CIMObjectPath->new($resultClass, $ns);
		$cop->setRefKeyValue('Element', $objectName);
		$cop->setRefKeyValue('Setting', $mcop);
		$result->handle($cop);
	}
	elsif ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_GLOBALSETTINGS))
	{
		die "Not implemented yet";
	}
	elsif ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_SHARESETTINGS))
	{
		die "Not implemented yet";
	}
}

#######################################################################
##                                                                   ##
##                 SHARE <--> SHARE SETTINGS PROVIDER                ##
##                                     (association class)           ##
##                                                                   ##
#######################################################################

package ShareShareSettings;
use base 'Config4GNU::WBEM::SimpleProvider';

sub associatorNames
{
	my $provider = shift;
	my ($env, $result, $ns, $objectName, $assocClass, $resultClass,
		$role, $resultRole) = @_;

	my $baseClass = $objectName->getClassName;
	if ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_SHARE))
	{
		my $shareName = $objectName->getStringKeyValue('Name');

		my $mcop = owperl::CIMObjectPath->new(
				main::CLASSNAME_SHARESETTINGS, $ns);
		$mcop->addStringKeyValue('ConfigFile', $Service::CONFIG_FILES[0]);
		$mcop->addStringKeyValue('Name', $shareName);
		$result->handle($mcop);
	}
	elsif ($provider->is_subclass($env, $ns, $baseClass, main::CLASSNAME_SHARESETTINGS))
	{
		die "Not implemented yet";
	}
}
